"""
A model for TorsionDrive
"""

import copy
import json
from typing import Any, Dict, List, Tuple, Union

from pydantic import BaseModel

from .common_models import QCMeta, Provenance, Molecule, json_encoders, hash_dictionary

__all__ = ["TorsionDriveInput", "TorsionDrive"]


class TorsionDriveInput(BaseModel):
    """
    A TorsionDrive Input base class
    """

    class TDOptions(BaseModel):
        """
        TorsionDrive options
        """
        dihedrals: List[Tuple[int, int, int, int]]
        grid_spacing: List[float]

        class Config:
            allow_extra = True
            allow_mutation = False

    class OptOptions(BaseModel):
        """
        TorsionDrive options
        """
        program: str

        class Config:
            allow_extra = True
            allow_mutation = False

    program: str = "torsiondrive"
    procedure: str = "torsiondrive"
    initial_molecule: List[Union[str, Molecule]]
    torsiondrive_meta: TDOptions
    optimization_meta: OptOptions
    qc_meta: QCMeta

    def __init__(self, **data):
        mol = data["initial_molecule"]
        if isinstance(mol, (str, dict, Molecule)):
            data["initial_molecule"] = [mol]

        BaseModel.__init__(self, **data)

    class Config:
        allow_mutation = False
        json_encoders = json_encoders


class TorsionDrive(TorsionDriveInput):
    """
    A interface to the raw JSON data of a TorsionDrive torsion scan run.
    """

    # Client and local data
    client: Any = None
    cache: Dict[str, Any] = {}

    # Identification
    id: str = None
    success: bool = False
    status: str = "INCOMPLETE"
    hash_index: str = None

    provenance: Provenance

    # Data pointers
    initial_molecule: List[str]
    final_energy_dict: Dict[str, float]
    optimization_history: Dict[str, List[str]]
    minimum_positions: Dict[str, int]

    class Config:
        allow_mutation = False
        json_encoders = json_encoders

    def __init__(self, **data):
        super().__init__(**data)

        # Set hash index if not present
        if self.hash_index is None:
            self.__values__["hash_index"] = self.get_hash_index()

    def __str__(self):
        """
        Simplified torsiondrive string representation.

        Returns
        -------
        ret : str
            A representation of the current TorsionDrive status.

        Examples
        --------

        >>> repr(torsiondrive_obj)
        TorsionDrive(id='5b7f1fd57b87872d2c5d0a6d', success=True, molecule_id='5b7f1fd57b87872d2c5d0a6c')
        """

        ret = "TorsionDrive("
        ret += "id='{}', ".format(self.id)
        ret += "success='{}', ".format(self.success)
        ret += "initial_molecule='{}')".format(self.initial_molecule)

        return ret

## Utility

    def _serialize_key(self, key):
        if isinstance(key, (int, float)):
            key = (int(key), )

        return json.dumps(key)

    def _deserialize_key(self, key):
        return tuple(json.loads(key))

    def dict(self, exclude=set(), include=None, by_alias=False):
        exclude |= {"client", "cache"}
        return super().dict(include=include, exclude=exclude, by_alias=by_alias)

    def json_dict(self):
        return json.loads(self.json())

    def get_hash_index(self):

        data = self.dict(
            include=["initial_molecule", "program", "procedure", "torsiondrive_meta", "optimization_meta", "qc_meta"])

        return hash_dictionary(data)

## Query

    def get_history(self):
        """Pulls all optimization trajectories to local data.

        Returns
        -------
        dict
            The optimization history
        """

        if "history" not in self.cache:

            # Grab procedures
            needed_ids = [x for v in self.optimization_history.values() for x in v]
            objects = self.client.get_procedures({"id": needed_ids})
            procedures = {v._id: v for v in objects}

            # Move procedures into the correct order
            ret = {}
            for key, hashes in self.optimization_history.items():
                tmp = []
                for h in hashes:
                    tmp.append(procedures[h])
                ret[key] = tmp

            self.cache["history"] = ret

        return self.cache["history"]

    def final_energies(self, key=None):
        """
        Provides the final optimized energies at each grid point.

        Parameters
        ----------
        key : None, optional
            Specifies a single entry to pull from.


        Returns
        -------
        energy : float, dict
            Returns energies at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------

        >>> torsiondrive_obj.final_energies()
        {(-90,): -148.7641654446243, (180,): -148.76501336993732, (0,): -148.75056290106735, (90,): -148.7641654446148}
        """

        if key is None:
            return {self._deserialize_key(k): v for k, v in self.final_energy_dict.items()}
        else:

            return self.final_energy_dict[self._serialize_key(key)]

    def final_molecules(self, key=None):
        """Returns the optimized molecules at each grid point

        Parameters
        ----------
        key : None, optional
            Specifies a single entry to pull from.


        Returns
        -------
        energy : dict
            Returns molecule at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------

        >>> torsiondrive_obj.final_energies()
        {(-90,):{'symbols': ['H', 'O', 'O', 'H'], 'geometry': [1.72669422, 1.28135788, ... }
        """

        if "final_molecules" not in self.cache:

            ret = {}
            for k, tasks in self.get_history().items():
                minpos = self.minimum_positions[k]

                ret[k] = tasks[minpos].final_molecule()

            self.cache["final_molecules"] = ret

        data = self.cache["final_molecules"]

        if key is None:
            return {self._deserialize_key(k): copy.deepcopy(v) for k, v in data.items()}
        else:

            return data[self._serialize_key(key)]



"""
Queue backend abstraction manager.
"""

import asyncio
import json
import logging
import socket
import uuid
from typing import Any, Callable, Dict, List, Optional

import qcengine
import tornado.ioloop

from qcfractal.extras import get_information
from .adapters import build_queue_adapter
from ..interface.data import get_molecule
from ..interface.models.rest_models import (
    QueueManagerGETBody, QueueManagerGETResponse, QueueManagerPOSTBody, QueueManagerPOSTResponse, QueueManagerPUTBody,
    QueueManagerPUTResponse
)

__all__ = ["QueueManager"]


class QueueManager:
    """
    This object maintains a computational queue and watches for finished tasks for different
    queue backends. Finished tasks are added to the database and removed from the queue.

    Attributes
    ----------
    client : FractalClient
        A Portal client to connect to a server
    queue_adapter : QueueAdapter
        The DBAdapter class for queue abstraction
    errors : dict
        A dictionary of current errors
    logger : logging.logger. Optional, Default: None
        A logger for the QueueManager
    """

    def __init__(self,
                 client: Any,
                 queue_client: Any,
                 loop: Any=None,
                 logger: Optional[logging.Logger]=None,
                 max_tasks: int=1000,
                 queue_tag: str=None,
                 cluster: str="unknown",
                 update_frequency: int=2,
                 verbose: bool=True,
                 cores_per_task: Optional[int] = None,
                 memory_per_task: Optional[int] = None):
        """
        Parameters
        ----------
        client : FractalClient
            A Portal client to connect to a server
        queue_client : QueueAdapter
            The DBAdapter class for queue abstraction
        storage_socket : DBSocket
            A socket for the backend database
        loop : IOLoop
            The running Tornado IOLoop
        logger : logging.Logger, Optional. Default: None
            A logger for the QueueManager
        max_tasks : int
            The maximum number of tasks to hold at any given time
        queue_tag : str
            Allows managers to pull from specific tags
        cluster : str
            The cluster the manager belongs to
        update_frequency : int
            The frequency to check for new tasks in seconds
        cores_per_task : int, optional, Default: None
            How many CPU cores per computation task to allocate for QCEngine
            None indicates "use however many you can detect"
        memory_per_task: int, optional, Default: None
            How much memory, in GiB, per computation task to allocate for QCEngine
            None indicates "use however much you can consume"
        """

        # Setup logging
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('QueueManager')

        self.name_data = {"cluster": cluster, "hostname": socket.gethostname(), "uuid": str(uuid.uuid4())}
        self._name = self.name_data["cluster"] + "-" + self.name_data["hostname"] + "-" + self.name_data["uuid"]

        self.client = client
        self.cores_per_task = cores_per_task
        self.memory_per_task = memory_per_task
        self.queue_adapter = build_queue_adapter(queue_client,
                                                 logger=self.logger,
                                                 cores_per_task=self.cores_per_task,
                                                 memory_per_task=self.memory_per_task)
        self.max_tasks = max_tasks
        self.queue_tag = queue_tag
        self.verbose = verbose

        self.update_frequency = update_frequency
        self.periodic = {}
        self.active = 0
        self.exit_callbacks = []

        # Pull the current loop if we need it
        self.loop = loop or tornado.ioloop.IOLoop.current()

        # Build a meta header
        meta_packet = self.name_data.copy()
        meta_packet["tag"] = self.queue_tag
        meta_packet["max_tasks"] = self.max_tasks
        self.meta_packet = json.dumps(meta_packet)

        self.logger.info("QueueManager:")
        self.logger.info("    Version:         {}\n".format(get_information("version")))

        if self.verbose:
            self.logger.info("    Name Information:")
            self.logger.info("        Cluster:     {}".format(self.name_data["cluster"]))
            self.logger.info("        Hostname:    {}".format(self.name_data["hostname"]))
            self.logger.info("        UUID:        {}\n".format(self.name_data["uuid"]))

        self.logger.info("    Queue Adapter:")
        self.logger.info("        {}\n".format(self.queue_adapter))

        if self.verbose:
            self.logger.info("    QCEngine:")
            self.logger.info("        Version:    {}\n".format(qcengine.__version__))

        # DGAS Note: Note super happy about how this if/else turned out. Looking for alternatives.
        if self.connected():
            # Pull server info
            self.server_info = client.server_information()
            self.server_name = self.server_info["name"]
            self.server_version = self.server_info["version"]
            self.heartbeat_frequency = self.server_info["heartbeat_frequency"]

            # Tell the server we are up and running
            payload = self._payload_template()
            payload["data"]["operation"] = "startup"
            put_body = QueueManagerPUTBody(**payload)
            r = self.client._request("put", "queue_manager", data=put_body.json())
            _ = QueueManagerPUTResponse.parse_raw(r.text)  # Validate

            if self.verbose:
                self.logger.info("    Connected:")
                self.logger.info("        Version:     {}".format(self.server_version))
                self.logger.info("        Address:     {}".format(self.client.address))
                self.logger.info("        Name:        {}".format(self.server_name))
                self.logger.info("        Queue tag:   {}".format(self.queue_tag))
                self.logger.info("        Username:    {}\n".format(self.client.username))

        else:
            self.logger.info("    QCFractal server information:")
            self.logger.info("        Not connected, some actions will not be available")

    def _payload_template(self):
        return {"meta": json.loads(self.meta_packet), "data": {}}

## Accessors

    def name(self) -> str:
        """
        Returns the Managers full name.
        """
        return self._name

    def connected(self) -> bool:
        """
        Checks the connection to the server.
        """
        return self.client is not None

    def assert_connected(self) -> None:
        """
        Raises an error for functions that require a server connection.
        """
        if self.connected() is False:
            raise AttributeError("Manager is not connected to a server, this operations is not available.")

## Start/stop functionality

    def start(self) -> None:
        """
        Starts up all IOLoops and processes
        """

        self.assert_connected()

        self.logger.info("QueueManager successfully started. Starting IOLoop.\n")

        # Add services callback, cb freq is given in milliseconds
        update = tornado.ioloop.PeriodicCallback(self.update, 1000 * self.update_frequency)
        update.start()
        self.periodic["update"] = update

        # Add heartbeat
        heartbeat_frequency = int(0.8 * 1000 * self.heartbeat_frequency)  # Beat at 80% of cutoff time
        heartbeat = tornado.ioloop.PeriodicCallback(self.heartbeat, heartbeat_frequency)
        heartbeat.start()
        self.periodic["heartbeat"] = heartbeat

        # Soft quit with a keyboard interrupt
        self.running = True
        self.loop.start()

    def stop(self) -> None:
        """
        Shuts down all IOLoops and periodic updates
        """

        # Push data back to the server
        self.shutdown()

        # Close down the adapter
        self.close_adapter()

        # Stop callbacks
        for cb in self.periodic.values():
            cb.stop()

        # Call exit callbacks
        for func, args, kwargs in self.exit_callbacks:
            func(*args, **kwargs)

        # Stop loop
        if not asyncio.get_event_loop().is_running():  # Only works on Py3
            self.loop.stop()

        self.loop.close(all_fds=True)
        self.logger.info("QueueManager stopping gracefully. Stopped IOLoop.\n")

    def close_adapter(self) -> bool:
        """
        Closes down the underlying adapter
        """

        return self.queue_adapter.close()

## Queue Manager functions

    def heartbeat(self) -> None:
        """
        Provides a heartbeat to the connected Server
        """

        self.assert_connected()

        payload = self._payload_template()
        payload["data"]["operation"] = "heartbeat"
        put_body = QueueManagerPUTBody(**payload)
        r = self.client._request("put", "queue_manager", data=put_body.json(), noraise=True)
        if r.status_code != 200:
            # TODO something as we didnt successfully add the data
            self.logger.warning("Heartbeat was not successful.")

        _ = QueueManagerPUTResponse.parse_raw(r.text)  # Validate

    def shutdown(self) -> Dict[str, Any]:
        """
        Shutdown the manager and returns tasks to queue.
        """
        self.assert_connected()

        payload = self._payload_template()
        payload["data"]["operation"] = "shutdown"
        put_body = QueueManagerPUTBody(**payload)
        r = self.client._request("put", "queue_manager", data=put_body.json(), noraise=True)

        if r.status_code != 200:
            # TODO something as we didnt successfully add the data
            self.logger.warning("Shutdown was not successful. This may delay queued tasks.")
            return {"nshutdown": 0}

        response = QueueManagerPUTResponse.parse_raw(r.text)
        nshutdown = response.data["nshutdown"]
        self.logger.info("Shutdown was successful, {} tasks returned to master queue.".format(nshutdown))
        return response.data

    def add_exit_callback(self, callback: Callable, *args: List[Any], **kwargs: Dict[Any, Any]) -> None:
        """Adds additional callbacks to perform when closing down the server

        Parameters
        ----------
        callback : callable
            The function to call at exit
        *args
            Arguements to call with the function.
        **kwargs
            Kwargs to call with the function.

        """
        self.exit_callbacks.append((callback, args, kwargs))

    def update(self, new_tasks: bool=True) -> bool:
        """Examines the queue for completed tasks and adds successful completions to the database
        while unsuccessful are logged for future inspection

        """

        self.assert_connected()

        results = self.queue_adapter.acquire_complete()
        if len(results):
            payload = self._payload_template()
            payload["data"] = results
            body = QueueManagerPOSTBody(**payload)
            r = self.client._request("post", "queue_manager", data=body.json(), noraise=True)
            if r.status_code != 200:
                # TODO something as we didnt successfully add the data
                self.logger.warning("Post complete tasks was not successful. Data may be lost.")

            _ = QueueManagerPOSTResponse.parse_raw(r.text)  # Ensure validation from server

            self.active -= len(results)

        open_slots = max(0, self.max_tasks - self.active)

        if (new_tasks is False) or (open_slots == 0):
            return True

        # Get new tasks
        payload = self._payload_template()
        payload["data"]["limit"] = open_slots
        body = QueueManagerGETBody(**payload)
        r = self.client._request("get", "queue_manager", data=body.json(), noraise=True)
        if r.status_code != 200:
            # TODO something as we didnt successfully get data
            self.logger.warning("Acquisition of new tasks was not successful.")

        response = QueueManagerGETResponse.parse_raw(r.text)
        new_tasks = response.data

        # Add new tasks to queue
        self.queue_adapter.submit_tasks(new_tasks)
        self.active += len(new_tasks)
        return True

    def await_results(self) -> bool:
        """A synchronous method for testing or small launches
        that awaits task completion.

        Returns
        -------
        bool
            Return True if the operation completed successfully
        """

        self.assert_connected()

        self.update()
        self.queue_adapter.await_results()
        self.update(new_tasks=False)
        return True

    def list_current_tasks(self) -> List[Any]:
        """Provides a list of tasks currently in the queue along
        with the associated keys

        Returns
        -------
        ret : list of tuples
            All tasks currently still in the database
        """
        return self.queue_adapter.list_tasks()

    def test(self) -> bool:
        """
        Tests all known programs with simple inputs to check if the Adapter is correctly instantiated.
        """

        from qcfractal import testing
        self.logger.info("Testing requested, generating tasks")
        task_base = json.dumps({
            "spec": {
                "function":
                "qcengine.compute",
                "args": [{
                    "molecule": get_molecule("hooh.json").json_dict(),
                    "driver": "energy",
                    "model": {},
                    "keywords": {},
                    "return_output": False
                }, "program"],
                "kwargs": {}
            },
            "parser": "single",
            "hooks": []
        })

        programs = {
            "rdkit": {
                "method": "UFF",
                "basis": None
            },
            "torchani": {
                "method": "ANI1",
                "basis": None
            },
            "psi4": {
                "method": "HF",
                "basis": "sto-3g"
            },
        }
        tasks = []
        found_programs = []

        for program, model in programs.items():
            if testing.has_module(program):
                self.logger.info("Found program {}, adding to testing queue.".format(program))
                found_programs.append(program)
            else:
                self.logger.warning("Could not find program {}, skipping tests.".format(program))
                continue

            task = json.loads(task_base)
            task["id"] = program
            task["spec"]["args"][0]["model"] = model
            task["spec"]["args"][1] = program

            tasks.append(task)

        self.queue_adapter.submit_tasks(tasks)

        self.logger.info("Testing tasks submitting, awaiting results.\n")
        self.queue_adapter.await_results()

        results = self.queue_adapter.acquire_complete()
        self.logger.info("Testing results acquired.")

        missing_programs = results.keys() - set(found_programs)
        if len(missing_programs):
            self.logger.error("Not all tasks were retrieved, missing programs {}.".format(missing_programs))
            raise ValueError("Testing failed, not all tasks were retrieved.")
        else:
            self.logger.info("All tasks retrieved successfully.")

        failures = 0
        for k, result in results.items():
            if result[0]["success"]:
                self.logger.info("  {} - PASSED".format(k))
            else:
                self.logger.error("  {} - FAILED!".format(k))
                failures += 1

        if failures:
            self.logger.error("{}/{} tasks failed!".format(failures, len(results)))
            return False
        else:
            self.logger.info("All tasks completed successfully!")
            return True



"""
Python init for QCPortal collections
"""

from .collection_utils import collection_factory
from .dataset import Dataset
from .reaction_dataset import ReactionDataset
from .generic import Generic
from .openffworkflow import OpenFFWorkflow



"""
Queue backend abstraction manager.
"""

import collections
import traceback

from .. import procedures
from .. import services
from ..interface.models.common_models import Molecule
from ..interface.models.rest_models import (
    TaskQueueGETBody, TaskQueueGETResponse, TaskQueuePOSTBody, TaskQueuePOSTResponse,
    ServiceQueueGETBody, ServiceQueueGETResponse, ServiceQueuePOSTBody, ServiceQueuePOSTResponse,
    QueueManagerGETBody, QueueManagerGETResponse, QueueManagerPOSTBody, QueueManagerPOSTResponse,
    QueueManagerPUTBody, QueueManagerPUTResponse
)  # yapf: disable
from ..web_handlers import APIHandler


class TaskQueueHandler(APIHandler):
    """
    Takes in a data packet the contains the molecule_hash, modelchem and options objects.
    """

    def post(self):
        """Summary
        """
        self.authenticate("compute")

        # Grab objects
        storage = self.objects["storage_socket"]

        post = TaskQueuePOSTBody.parse_raw(self.request.body)

        # Format and submit tasks
        procedure_parser = procedures.get_procedure_parser(post.meta["procedure"], storage)
        payload = procedure_parser.submit_tasks(post)

        response = TaskQueuePOSTResponse(**payload)
        self.logger.info("TaskQueue: Added {} tasks.".format(response.meta.n_inserted))

        self.write(response.json())

    def get(self):
        """Posts new services to the service queue
        """
        self.authenticate("read")

        # Grab objects
        storage = self.objects["storage_socket"]

        body = TaskQueueGETBody.parse_raw(self.request.body)

        tasks = storage.get_queue(**body.data, projection=body.meta.projection)
        response = TaskQueueGETResponse(**tasks)

        self.write(response.json())


class ServiceQueueHandler(APIHandler):
    """
    Takes in a data packet the contains the molecule_hash, modelchem and options objects.
    """

    def post(self):
        """Posts new services to the service queue
        """
        self.authenticate("compute")

        # Grab objects
        storage = self.objects["storage_socket"]

        # Figure out initial molecules
        service_input = ServiceQueuePOSTBody.parse_raw(self.request.body).data
        meta = {
            "validation_errors": [],
            "duplicates": [],
            "n_inserted": 0,
            "success": True,
            "errors": [],
            "error_description": ""
        }

        # Get molecules with ids
        if isinstance(service_input.initial_molecule, list):
            mol_query = storage.get_add_molecules_mixed(service_input.initial_molecule)
            molecules = [Molecule(**mol) for mol in mol_query["data"]]
            if len(molecules) != len(service_input.initial_molecule):
                raise KeyError("We should catch this error.")
        else:
            mol_query = storage.get_add_molecules_mixed([service_input.initial_molecule])
            molecules = Molecule(**mol_query["data"][0])

        # Update the input and build a service object
        service_input = service_input.copy(update={"initial_molecule": molecules})
        new_service = services.initializer(storage, service_input)

        data = {"hash_index": new_service.hash_index, "status": None}

        # Figure out complete services
        duplicate = storage.get_procedures_by_id(
            hash_index=new_service.hash_index, projection={"hash_index": True})["data"]
        if duplicate:
            data["status"] = "completed"
            # meta["duplicates"] = duplicate
        else:
            # Add services to the database
            ret = storage.add_services([new_service.json_dict()])
            if ret["meta"]["duplicates"]:
                data["status"] = "running"
                meta["n_inserted"] = 1
            else:
                data["status"] = "submitted"

        response = ServiceQueuePOSTResponse(data=data, meta=meta)
        self.logger.info("ServiceQueue: Added {} services.\n".format(response.meta.n_inserted))

        self.write(response.json())

    def get(self):
        """Gets services from the service queue
        """
        self.authenticate("read")

        # Grab objects
        storage = self.objects["storage_socket"]

        body = ServiceQueueGETBody.parse_raw(self.request.body)

        projection = {x: True for x in ["status", "error_message", "tag"]}
        ret = storage.get_services(**body.data, projection=projection)
        response = ServiceQueueGETResponse(**ret)

        self.write(response.json())


class QueueManagerHandler(APIHandler):
    """
    Takes in a data packet the contains the molecule_hash, modelchem and options objects.
    Manages the external
    """

    def _get_name_from_metadata(self, meta):
        """
        Form the canonical name string.
        """
        ret = meta.cluster + "-" + meta.hostname + "-" + meta.uuid
        return ret

    @staticmethod
    def insert_complete_tasks(storage_socket, results, logger):
        # Pivot data so that we group all results in categories
        new_results = collections.defaultdict(list)
        error_data = []

        task_success = 0
        task_failures = 0
        task_totals = len(results.items())
        for key, (result, parser, hooks) in results.items():
            try:

                # Successful task
                if result["success"] is True:
                    result["task_id"] = key
                    new_results[parser].append((result, hooks))
                    task_success += 1

                # Failed task
                else:
                    if "error" in result:
                        logger.warning(
                            "Found old-style error field, please change to 'error_message'. Will be deprecated")
                        error = result["error"]
                        result["error_message"] = error
                    elif "error_message" in result:
                        error = result["error_message"]
                    else:
                        error = "No error supplied"

                    logger.info("Computation key did not complete successfully:\n\t{}\n"
                                "Because: {}".format(str(key), error["error_message"]))

                    error_data.append((key, error))
                    task_failures += 1
            except Exception as e:
                msg = "Internal FractalServer Error:\n" + traceback.format_exc()
                logger.info("update: ERROR\n{}".format(msg))
                error_data.append((key, msg))
                task_failures += 1

        if task_totals:
            logger.info("QueueManager: Found {} complete tasks ({} successful, {} failed).".format(
                task_totals, task_success, task_failures))

        # Run output parsers
        completed = []
        hooks = []
        for k, v in new_results.items():  # todo: can be merged? do they have diff k?
            procedure_parser = procedures.get_procedure_parser(k, storage_socket)
            com, err, hks = procedure_parser.parse_output(v)
            completed.extend(com)
            error_data.extend(err)
            hooks.extend(hks)

        # Handle hooks and complete tasks
        storage_socket.handle_hooks(hooks)
        storage_socket.queue_mark_complete(completed)
        storage_socket.queue_mark_error(error_data)
        return len(completed), len(error_data)

    def get(self):
        """Pulls new tasks from the Servers queue
        """
        self.authenticate("queue")

        # Grab objects
        storage = self.objects["storage_socket"]

        body = QueueManagerGETBody.parse_raw(self.request.body)

        # Figure out metadata and kwargs
        name = self._get_name_from_metadata(body.meta)
        queue_tags = {
            "limit": body.data.limit,
            "tag": body.meta.tag,
        }  # yapf: disable

        # Grab new tasks and write out
        new_tasks = storage.queue_get_next(name, **queue_tags)
        response = QueueManagerGETResponse(
            meta={"n_found": len(new_tasks),
                  "success": True,
                  "errors": [],
                  "error_description": "",
                  "missing": []},
            data=new_tasks)
        self.write(response.json())
        self.logger.info("QueueManager: Served {} tasks.".format(response.meta.n_found))

        # Update manager logs
        storage.manager_update(name, submitted=len(new_tasks), **body.meta.dict())

    def post(self):
        """Posts complete tasks to the Servers queue
        """
        self.authenticate("queue")

        # Grab objects
        storage = self.objects["storage_socket"]

        body = QueueManagerPOSTBody.parse_raw(self.request.body)
        ret = self.insert_complete_tasks(storage, body.data, self.logger)

        response = QueueManagerPOSTResponse(
            meta={
                "n_inserted": ret[0],
                "duplicates": [],
                "validation_errors": [],
                "success": True,
                "errors": [],
                "error_description": "" if not ret[1] else "{} errors".format(ret[1])
            },
            data=True)
        self.write(response.json())
        self.logger.info("QueueManager: Acquired {} complete tasks.".format(len(body.data)))

        # Update manager logs
        name = self._get_name_from_metadata(body.meta)
        storage.manager_update(name, completed=len(body.data), **body.meta.dict())

    def put(self):
        """
        Various manager manipulation operations
        """
        self.authenticate("queue")

        storage = self.objects["storage_socket"]
        ret = True

        body = QueueManagerPUTBody.parse_raw(self.request.body)
        name = self._get_name_from_metadata(body.meta)
        op = body.data.operation
        if op == "startup":
            storage.manager_update(name, status="ACTIVE", **body.meta.dict())
            self.logger.info("QueueManager: New active manager {} detected.".format(name))

        elif op == "shutdown":
            nshutdown = storage.queue_reset_status(name)
            storage.manager_update(name, returned=nshutdown, status="INACTIVE", **body.meta.dict())

            self.logger.info("QueueManager: Shutdown of manager {} detected, recycling {} incomplete tasks.".format(
                name, nshutdown))

            ret = {"nshutdown": nshutdown}

        elif op == "heartbeat":
            storage.manager_update(name, status="ACTIVE", **body.meta.dict())
            self.logger.info("QueueManager: Heartbeat of manager {} detected.".format(name))

        else:
            msg = "Operation '{}' not understood.".format(op)
            from tornado.web import HTTPError
            raise HTTPError(status_code=400, reason=msg)

        response = QueueManagerPUTResponse(meta={}, data=ret)
        self.write(response.json())

        # Update manager logs



# def bench(factory, X, Y, X_test, Y_test, ref_coef):
#     gc.collect()
#
#     # start time
#     tstart = time()
#     clf = factory(alpha=alpha).fit(X, Y)
#     delta = (time() - tstart)
#     # stop time
#
#     print("duration: %0.3fs" % delta)
#     print("rmse: %f" % rmse(Y_test, clf.predict(X_test)))
#     print("mean coef abs diff: %f" % abs(ref_coef - clf.coef_.ravel()).mean())
#     return delta
#
#
# if __name__ == '__main__':
#     # Delayed import of matplotlib.pyplot
#     import matplotlib.pyplot as plt
#
#     scikit_results = []
#
#     n = 20
#     step = 500
#     n_features = 1000
#     n_informative = int(n_features / 10)
#     n_test_samples = 1000
#     for i in range(1, n + 1):
#         print('==================')
#         print('Iteration %s of %s' % (i, n))
#         print('==================')
#
#         X, Y, coef_ = make_regression(
#             n_samples=(i * step) + n_test_samples, n_features=n_features,
#             noise=0.1, n_informative=n_informative, coef=True)
#
#         X_test = X[-n_test_samples:]
#         Y_test = Y[-n_test_samples:]
#         X = X[:(i * step)]
#         Y = Y[:(i * step)]
#
#         print("benchmarking scikit-learn: ")
#         scikit_results.append(bench(ScikitLasso, X, Y, X_test, Y_test, coef_))
#         print("benchmarking glmnet: ")
#         # glmnet_results.append(bench(GlmnetLasso, X, Y, X_test, Y_test, coef_))
#
#     plt.clf()
#     xx = range(0, n * step, step)
#     plt.title('Lasso regression on sample dataset (%d features)' % n_features)
#     plt.plot(xx, scikit_results, 'b-', label='scikit-learn')
#     # plt.plot(xx, glmnet_results, 'r-', label='glmnet')
#     plt.legend()
#     plt.xlabel('number of samples to classify')
#     plt.ylabel('Time (s)')
#     plt.show()
#
#     # now do a benchmark where the number of points is fixed
#     # and the variable is the number of features
#
#     scikit_results = []
#     # glmnet_results = []
#     n = 20
#     step = 100
#     n_samples = 500
#
#     for i in range(1, n + 1):
#         print('==================')
#         print('Iteration %02d of %02d' % (i, n))
#         print('==================')
#         n_features = i * step
#         n_informative = n_features / 10
#
#         X, Y, coef_ = make_regression(
#             n_samples=(i * step) + n_test_samples, n_features=n_features,
#             noise=0.1, n_informative=n_informative, coef=True)
#
#         X_test = X[-n_test_samples:]
#         Y_test = Y[-n_test_samples:]
#         X = X[:n_samples]
#         Y = Y[:n_samples]
#
#         print("benchmarking scikit-learn: ")
#         scikit_results.append(bench(ScikitLasso, X, Y, X_test, Y_test, coef_))
#         print("benchmarking glmnet: ")
#         # glmnet_results.append(bench(GlmnetLasso, X, Y, X_test, Y_test, coef_))
#
#     xx = np.arange(100, 100 + n * step, step)
#     plt.figure('scikit-learn vs. glmnet benchmark results')
#     plt.title('Regression in high dimensional spaces (%d samples)' % n_samples)
#     plt.plot(xx, scikit_results, 'b-', label='scikit-learn')
#     # plt.plot(xx, glmnet_results, 'r-', label='glmnet')
#     plt.legend()
#     plt.xlabel('number of features')
#     plt.ylabel('Time (s)')
#     plt.axis('tight')
#     plt.show()


"""
Common models for QCPortal/Fractal
"""
import hashlib
import json
import numpy as np
from pydantic import BaseModel, validator
from typing import Any, Dict, Optional

from qcelemental.models import Molecule, Provenance, Result, ResultInput

__all__ = ["QCMeta", "json_encoders", "hash_dictionary", "KeywordSet"]

# Add in QCElemental models
__all__.extend(["Molecule", "Provenance"])

json_encoders = {np.ndarray: lambda v: v.flatten().tolist()}


def recursive_hash_prep(value: Any, **kwargs: Dict[str, Any]) -> Any:
    """
    Prepare a structure for hashing by lowercasing all values and round all floats
    """
    digits = kwargs.get("digits", 10)
    lowercase = kwargs.get("lowercase", True)

    if isinstance(value, (int, type(None))):
        pass

    elif isinstance(value, str):
        if lowercase:
            value = value.lower()

    elif isinstance(value, (list, tuple)):
        value = [recursive_hash_prep(x, **kwargs) for x in value]

    elif isinstance(value, dict):
        ret = {}
        for k, v in value.items():
            if lowercase:
                k = k.lower()
            ret[k] = recursive_hash_prep(v, **kwargs)
        value = ret

    elif isinstance(value, float):
        if digits:
            value = round(value, digits)
            if value == 0.0:  # Values rounded to zero
                value = 0

    else:
        raise TypeError("Invalid type in KeywordSet ({type(value)}), only simply Python types are allowed.")

    return value


def hash_dictionary(data: Dict[str, Any]) -> str:
    m = hashlib.sha1()
    m.update(json.dumps(data, sort_keys=True).encode("UTF-8"))
    return m.hexdigest()


class QCMeta(BaseModel):
    """
    The basic quantum chemistry meta specification
    """
    driver: str
    method: str
    basis: Optional[str] = None
    keywords: Optional[str] = None
    program: str


class KeywordSet(BaseModel):
    """
    An options object for the QCArchive ecosystem
    """
    id: Optional[str] = None
    program: str
    hash_index: str
    values: Dict[str, Any]
    lowercase: bool = True
    exact_floats: bool = False

    class Config:
        allow_extra = True
        allow_mutation = False

    def __init__(self, **data):

        build_index = False
        if ("hash_index" not in data) or data.pop("build_index", False):
            build_index = True
            data["hash_index"] = "placeholder"

        BaseModel.__init__(self, **data)

        # Overwrite options with massaged values
        kwargs = {"lowercase": self.lowercase}
        if self.exact_floats:
            kwargs["digits"] = False

        self.__values__["values"] = recursive_hash_prep(self.values, **kwargs)

        # Build a hash index if we need it
        if build_index:
            self.__values__["hash_index"] = self.get_hash_index()

    @validator('program')
    def check_program(cls, v):
        return v.lower()

    def get_hash_index(self):
        packet = self.values.copy()
        packet["program"] = self.program
        return hash_dictionary(packet)

    def json_dict(self, *args, **kwargs):
        return json.loads(self.json(*args, **kwargs))



"""
Explicit tests for queue manipulation.
"""

import pytest

import qcfractal.interface as portal
from qcfractal import testing, QueueManager
from qcfractal.testing import (
    reset_server_database, managed_compute_server, adapter_client_fixture)


@testing.using_rdkit
def test_adapter_single(managed_compute_server):
    client, server, manager = managed_compute_server
    reset_server_database(server)

    # Add compute
    hooh = portal.data.get_molecule("hooh.json")
    ret = client.add_compute("rdkit", "UFF", "", "energy", None, [hooh.json_dict()], tag="other")

    # Force manager compute and get results
    manager.await_results()
    ret = client.get_results()
    assert len(ret) == 1


@pytest.mark.parametrize("cores_per_task,memory_per_task", [
    (None, None),
    (1, 1),
    (2, 1.9)
])
@testing.using_psi4
def test_keyword_args_passing(adapter_client_fixture, cores_per_task, memory_per_task):
    psi4_mem_buffer = 0.95  # Memory consumption buffer on psi4
    adapter_client = adapter_client_fixture
    task_id = "123456789012345678901234"  # Placeholder ID
    tasks = [  # Emulate the QueueManager test function
        {
            "id": task_id,
            "spec": {
                "function":
                    "qcengine.compute",
                "args": [{
                    "molecule": portal.data.get_molecule("hooh.json").json(as_dict=True),
                    "driver": "energy",
                    "model": {"method": "HF",
                              "basis": "sto-3g"},
                    "keywords": {},
                    "return_output": True,
                    'qcfractal_tags': {'program': 'psi4', 'options': None}
                }, "psi4"],
                "kwargs": {}
            },
            "parser": "single",
            "hooks": [],
            "tag": "other"
        }
    ]
    # Spin up a test queue manager
    manager = QueueManager(None, adapter_client, cores_per_task=cores_per_task, memory_per_task=memory_per_task)
    # Operate on the adapter since there is no backend QCF Client
    manager.queue_adapter.submit_tasks(tasks)
    manager.queue_adapter.await_results()
    ret = manager.queue_adapter.acquire_complete()
    assert len(ret) == 1
    provenance = ret[task_id][0]['provenance']
    if cores_per_task is not None:
        assert provenance['nthreads'] == cores_per_task
    if memory_per_task is not None:
        assert provenance['memory'] == pytest.approx(memory_per_task*psi4_mem_buffer)


@testing.using_rdkit
def test_adapter_error_message(managed_compute_server):
    client, server, manager = managed_compute_server
    reset_server_database(server)

    # HOOH without connectivity, RDKit should fail
    hooh = portal.data.get_molecule("hooh.json").json_dict()
    del hooh["connectivity"]
    mol_ret = client.add_molecules({"hooh": hooh})

    ret = client.add_compute("rdkit", "UFF", "", "energy", None, mol_ret["hooh"])
    queue_id = ret.submitted[0]

    # Nothing should have happened yet
    assert len(manager.list_current_tasks()) == 0

    # Pull out a special iteration on the queue manager
    manager.update()
    assert len(manager.list_current_tasks()) == 1

    manager.await_results()
    assert len(manager.list_current_tasks()) == 0

    db = server.objects["storage_socket"]
    ret = db.get_queue(status="ERROR")["data"]

    assert len(ret) == 1
    assert "connectivity graph" in ret[0]["error"]["error_message"]
    server.objects["storage_socket"].queue_mark_complete([queue_id])


@testing.using_rdkit
def test_adapter_raised_error(managed_compute_server):
    client, server, manager = managed_compute_server
    reset_server_database(server)

    # HOOH without connectivity, RDKit should fail
    hooh = portal.data.get_molecule("hooh.json").json_dict()
    del hooh["connectivity"]
    mol_ret = client.add_molecules({"hooh": hooh})

    ret = client.add_compute("something_bad", "UFF", "", "energy", None, mol_ret["hooh"])
    queue_id = ret.submitted[0]

    manager.await_results()

    db = server.objects["storage_socket"]
    ret = db.get_queue(status="ERROR")["data"]

    assert len(ret) == 1
    assert "QCEngine Call Error" in ret[0]["error"]["error_message"]
    server.objects["storage_socket"].queue_mark_complete([queue_id])



"""
Utilities for CLI programs
"""

import copy
import importlib
import json
import signal

import yaml


def import_module(module):
    """Protected import of a module
    """
    try:
        ret = importlib.import_module(module)
    except ImportError:
        raise ImportError("Requested module '{}' not found.".format(module))

    return ret


def read_config_file(fname):
    """Reads a JSON or YAML file.
    """
    if fname.endswith(".yaml") or fname.endswith(".yml"):
        rfunc = yaml.load
    elif fname.endswith(".json"):
        rfunc = json.load
    else:
        raise TypeError("Did not understand file type {}.".format(fname))

    with open(fname, "r") as handle:
        ret = rfunc(handle)

    return ret


def argparse_config_merge(parser, parsed_options, config_options, parser_default=None):
    """Merges options between a configuration file and a parser

    Parameters
    ----------
    parser : ArgumentParser
    config_options : dict
    parser_default : None, optional
    """
    config_options = copy.deepcopy(config_options)

    default_options = vars(parser.parse_args(args=parser_default))
    diff = config_options.keys() - default_options.keys()
    if diff:
        raise argparse.ArgumentError(None,
                                     "Unknown arguments found in configuration file: {}.".format(", ".join(diff)))

    # Add in parsed options
    for k, v in parsed_options.items():

        # User has overridden default options, update config
        if v != parser.get_default(k):
            config_options[k] = v

        # Add in missing defaults
        if v not in config_options:
            config_options[k] = v

    return config_options


def install_signal_handlers(loop, cleanup):
    """
    Install cleanup handlers to shutdown on:
     - Keyboard Interupt (SIGINT)
     - Shutdown kill/pkill (SIGTERM)
    """

    old_handlers = {}

    def handle_signal(sig, frame):
        async def cleanup_and_stop():
            try:
                cleanup()
            finally:
                loop.stop()

        loop.add_callback_from_signal(cleanup_and_stop)

        # Add old handlers back in so we do not cleanup twice
        signal.signal(sig, old_handlers[sig])

    for sig in [signal.SIGINT, signal.SIGTERM]:
        old_handlers[sig] = signal.signal(sig, handle_signal)



"""
The json-schema for the Molecule definition
"""

molecule_schema = {
    "$schema":
    "http://json-schema.org/draft-04/schema#",
    "properties": {
        "symbols": {
            "description": "The atom symbol for each atom in the molecule.",
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "geometry": {
            "description": "The 3N XYZ coordinates of the atoms involved.",
            "type": "array",
            "items": {
                "type": "number"
            }
        },
        "masses": {
            "description": "The masses of the atoms in the molecule, canonical weights assumed if not given.",
            "type": "array",
            "items": {
                "type": "number"
            }
        },
        "name": {
            "description": "The name of the molecule.",
            "type": "string"
        },
        "identifiers": {
            "description": "Canonical chemical identifiers.",
            "type": "object",
            "properties": {
                "molecule_hash": {
                    "description": "A unique hash for molecules specific to QCFractal.",
                    "type": "string"
                },
                "molecular_formula": {
                    "description": "A string giving the symbol and symbol count for the molecule.",
                    "type": "string"
                },
                "smiles": {
                    "description": "Simplified Molecular Input Line Entry System line notation.",
                    "type": "string"
                },
                "inchi": {
                    "description": "IUPAC International Chemical Identifier line notation.",
                    "type": "string"
                },
                "inchikey": {
                    "description": "A SHA1 hash of the inchi description.",
                    "type": "string"
                },
            }
        },
        "comment": {
            "description": "Any additional comment one would attach to the molecule.",
            "type": "string"
        },
        "molecular_charge": {
            "description": "The overall charge of the molecule.",
            "type": "number",
            "default": 0.0
        },
        "molecular_multiplicity": {
            "description": "The overall multiplicity of the molecule.",
            "type": "number",
            "multipleOf": 1.0,
            "default": 1
        },
        "real": {
            "description": "A list describing if the atoms are real or ghost.",
            "type": "array",
            "items": {
                "type": "boolean"
            }
        },
        "connectivity": {
            "description":
            "(nat, ) list describing bonds within a molecule. Each element is a (atom1, atom2, order) tuple.",
            "type":
            "array",
            "items": {
                "type": "array",
                "minItems": 3,
                "maxItems": 3,
                "items": {
                    "type": "number",
                    "minimum": 0,
                }
            }
        },
        "fragments": {
            "description": "A list of indices (0-indexed) for molecular fragments within the topology.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "number",
                    "multipleOf": 1.0
                }
            }
        },
        "fragment_charges": {
            "description": "A list of charges associated with the fragments tuple.",
            "type": "array",
            "items": {
                "type": "number"
            }
        },
        "fragment_multiplicities": {
            "description": "A list of multiplicites associated with each fragment.",
            "type": "array",
            "items": {
                "type": "number",
                "multipleOf": 1.0
            }
        },
        "fix_com": {
            "description": "Whether to adjust to the molecule to the COM or not.",
            "type": "boolean",
            "default": False
        },
        "fix_orientation": {
            "description": "Whether to rotate the molecule to a standard orientation or not.",
            "type": "boolean",
            "default": False
        },
        "provenance": {
            "type": "object",
            "$ref": "#/definitions/provenance"
        }
    },
    "required": ["symbols", "geometry"],
    "description":
    "The physical cartesian representation of the molecular system",
    "definitions": {},
    "additionalProperties":
    False,

    # Custom components
    "hash_fields": [
        "symbols", "masses", "molecular_charge", "molecular_multiplicity", "real", "geometry", "fragments", "fragment_charges",
        "fragment_multiplicities", "connectivity"
    ],
    "required_definitions": ["provenance"]
}



import json

import pytest

from ..common_models import KeywordSet


@pytest.mark.parametrize("data, hash_index", [

    # Simple checks
    ({"program": "psi4", "values": {"hi": 5}},
     "c9ae3acb782cad09c6867d8433a2963fe7b724fd"),

    ({"program": "PSI4", "values": {"HI": 5}},
     "c9ae3acb782cad09c6867d8433a2963fe7b724fd"),

    # Check options combinations
    ({"program": "PSI4", "values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}},
     "2d93c6c81466a2c51ec2e1a4f9e53a2c6bd9e3ed"),

    ({"program": "PSI4", "values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}, "lowercase": False},
     "3a6a8e2fb0449ec3db81191daf5de66f9a63c088"),

    ({"program": "PSI4", "values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}, "exact_floats": True},
     "7860429a00fe05f4e3a829c2e5680cd5a0fa94b6"),

    ({"program": "PSI4", "values": {"HI": [1, 2, 3], "All": [1.e-5, 2.e-3, 1.e-16]}, "exact_floats": True, "lowercase": False},
     "f0c58a301466f29c43cc303260e044e94551bc0a"),

    # Check recursive
    ({"program": "PSI4", "values": {"d1": {"D2": [1.e-17, 5]}}},
     "e50f3c0a580ca846b484a21a5a28058785fb8d00"),

    ({"program": "PSI4", "values": {"d1": {"D2": [1.e-17, 5]}}, "exact_floats": True, "lowercase": False},
     "ab6ce6dade7454730d2872c95e8f84449e00b321"),

    # Check hash_index build
    ({"program": "PSI4", "values": {}, "hash_index": "waffles"},
     "waffles"),

    ({"program": "PSI4", "values": {}, "hash_index": "waffles", "build_index": True},
     "10bdc0b99338761032d9d0f999ede01295185e5d"),
]) # yapf: disable
def test_option_canary_hash(data, hash_index):
    """
    Canary test aginst possible changes in the options hash_index
    """

    opt = KeywordSet(**data)

    assert hash_index == opt.hash_index, opt.keywords


@pytest.mark.parametrize("data1, data2", [
    # Test case sensitivity
    ({"program": "Psi4", "values": {}},
     {"program": "psi4", "values": {}}),

    # Test float creation
    ({"program": "Psi4", "values": {"CAPS": 5, "other": 4.e-3}},
     {"program": "psi4", "values": {"caps": 5, "other": 0.004}}),

    ({"program": "Psi4", "values": {"other": 4.123e-5}},
     {"program": "psi4", "values": {"OTHER": 0.00004123}}),

    # Test list of floats
    ({"program": "S", "values": {"other": [1.11e-2, 2.22e-3]}},
     {"program": "s", "values": {"other": [0.0111, 0.00222]}}),

    # Test small floats
    ({"program": "S", "values": {"other": 1 + 1.e-17}},
     {"program": "s", "values": {"other": 1.0 - 1.e-17}}),

    # Test dict order
    ({"program": "S", "values": {"a": 5, "b": 6, "c": None}},
     {"program": "s", "values": {"b": 6, "a": 5, "c": None}}),

    # Check recusive
    ({"program": "PSI4", "values": {"d1": {"D2": [0.0, 5], "d3": (3, 1.e-17)}}},
     {"program": "PSI4", "values": {"d1": {"d2": [1.e-17, 5], "d3": (3, 0)}}}),

]) # yapf: disable
def test_option_comparison_hash(data1, data2):
    """
    Ensure the hash_index finds collisions correctly before and after serialization.
    """
    opt1 = KeywordSet(**data1)
    opt2 = KeywordSet(**data2)

    # Check after serialization and rebuild index
    opt1s = KeywordSet(**json.loads(opt1.json()), build_index=True)
    opt2s = KeywordSet(**json.loads(opt2.json()), build_index=True)

    # Paranoid, try all combinations
    assert opt1.hash_index == opt2.hash_index
    assert opt1.hash_index == opt2s.hash_index
    assert opt1s.hash_index == opt2.hash_index
    assert opt1s.hash_index == opt2s.hash_index


"""
Misc information and runtime information.
"""

from . import _version

__all__ = ["get_information"]

versions = _version.get_versions()

__info = {"version": versions['version'], "git_revision": versions['full-revisionid']}


def get_information(key):
    """
    Obtains a variety of runtime information about QCFractal.
    """
    key = key.lower()
    if key not in __info:
        raise KeyError("Information key '{}' not understood.".format(key))

    return __info[key]



"""
Mongoengine Database class to handle access to mongoDB through ODM
"""

try:
    import pymongo
except ImportError:
    raise ImportError(
        "Mongostorage_socket requires pymongo, please install this python module or try a different db_socket.")

try:
    import mongoengine
except ImportError:
    raise ImportError(
        "Mongoengine_socket requires mongoengine, please install this python module or try a different db_socket.")

import collections
import json
import logging
from datetime import datetime as dt
from typing import List, Union, Dict, Sequence

import bcrypt
import bson.errors
import mongoengine as db
import mongoengine.errors
from bson.objectid import ObjectId
# import models
from mongoengine.connection import disconnect, get_db

from qcfractal.storage_sockets.models import Keywords, Collection, Result, \
    TaskQueue, Procedure, User, Molecule, QueueManager, ServiceQueue
from . import storage_utils
# Pull in the hashing algorithms from the client
from .. import interface


# from bson.dbref import DBRef


def _translate_id_index(index):
    if index in ["id", "ids"]:
        return "_id"
    else:
        raise KeyError("Id Index alias '{}' not understood".format(index))


def _str_to_indices(ids):
    for num, x in enumerate(ids):
        if isinstance(x, str):
            ids[num] = ObjectId(x)


def _str_to_indices_with_errors(ids):
    if isinstance(ids, (str, ObjectId)):
        ids = [ids]

    good = []
    bad = []
    for x in ids:
        if isinstance(x, str):
            try:
                good.append(ObjectId(x))
            except bson.errors.InvalidId:
                bad.append(x)
        elif isinstance(x, ObjectId):
            good.append(x)
        else:
            bad.append(x)
    return good, bad


class MongoengineSocket:
    """
        Mongoengine QCDB wrapper class.
    """

    def __init__(self,
                 uri,
                 project="molssidb",
                 bypass_security=False,
                 authMechanism="SCRAM-SHA-1",
                 authSource=None,
                 logger=None,
                 max_limit=1000):
        """
        Constructs a new socket where url and port points towards a Mongod instance.

        """

        # Logging data
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('MongoengineSocket')

        # Security
        self._bypass_security = bypass_security

        # Important: this dict is Not used for creating indices
        # To be removed and replaced by ME functions
        self._table_indices = {
            "collection": interface.schema.get_table_indices("collection"),
            "options": interface.schema.get_table_indices("options"),
            "result": interface.schema.get_table_indices("result"),
            "molecule": interface.schema.get_table_indices("molecule"),
            "procedure": interface.schema.get_table_indices("procedure"),
            "service_queue": interface.schema.get_table_indices("service_queue"),
            "task_queue": interface.schema.get_table_indices("task_queue"),
            "user": ("username", ),
            "queue_manager": ("name", )
        }
        # self._valid_tables = set(self._table_indices.keys())
        # self._table_unique_indices = {
        #     "collections": True,
        #     "options": True,
        #     "results": True,
        #     "molecules": False,
        #     "procedures": False,
        #     "service_queue": False,
        #     "task_queue": False,
        #     "users": True,
        #     "queue_managers": True,
        # }

        self._lower_results_index = ["method", "basis", "keywords", "program"]

        # disconnect from any active default connection
        disconnect()

        # Build MongoClient
        expanded_uri = pymongo.uri_parser.parse_uri(uri)
        if expanded_uri["password"] is not None:
            # connect to mongoengine
            self.client = db.connect(db=project, host=uri, authMechanism=authMechanism, authSource=authSource)
        else:
            # connect to mongoengine
            self.client = db.connect(db=project, host=uri)

        self._url, self._port = expanded_uri["nodelist"][0]

        try:
            version_array = self.client.server_info()['versionArray']

            if tuple(version_array) < (3, 2):
                raise RuntimeError
        except AttributeError:
            raise RuntimeError(
                "Could not detect MongoDB version at URL {}. It may be a very old version or installed incorrectly. "
                "Choosing to stop instead of assuming version is at least 3.2.".format(uri))
        except RuntimeError:
            # Trap low version
            raise RuntimeError("Connected MongoDB at URL {} needs to be at least version 3.2, found version {}.".
                               format(uri, self.client.server_info()['version']))

        # Isolate objects to this single project DB
        self._project_name = project
        self._tables = self.client[project]
        self._max_limit = max_limit

    ### Mongo meta functions

    def __str__(self):
        return "<MongoSocket: address='{0:s}:{1:d}:{2:s}'>".format(str(self._url), self._port, str(self._tables_name))

    def _clear_db(self, db_name: str):
        """Dangerous, make sure you are deleting the right DB"""

        logging.warning("Clearing database '{}' and dropping all tables.".format(db_name))

        # make sure it's the right DB
        if get_db().name == db_name:
            logging.info('Clearing database: {}'.format(db_name))
            Result.drop_collection()
            Molecule.drop_collection()
            Keywords.drop_collection()
            Collection.drop_collection()
            TaskQueue.drop_collection()
            QueueManager.drop_collection()
            Procedure.drop_collection()
            User.drop_collection()

            self.client.drop_database(db_name)

        QueueManager.objects(name='').first()  # init

    def get_project_name(self):
        return self._project_name

    def get_add_molecules_mixed(self, data):
        """
        Get or add the given molecules (if they don't exit).
        Molecules are given in a mixed format, either as a dict of mol data
        or as existing mol id

        TODO: to be split into get by_id and get_by_data
        """

        meta = storage_utils.get_metadata()

        ordered_mol_dict = {indx: mol for indx, mol in enumerate(data)}
        dict_mols = {}
        id_mols = {}
        for idx, mol in ordered_mol_dict.items():
            if isinstance(mol, str):
                id_mols[idx] = mol
            elif isinstance(mol, dict):
                mol.pop("id", None)
                dict_mols[idx] = mol
            elif isinstance(mol, interface.models.common_models.Molecule):
                mol_json = mol.json_dict()
                mol_json.pop("id", None)
                dict_mols[idx] = mol_json
            else:
                meta["errors"].append((idx, "Data type not understood"))

        ret_mols = {}

        # Add all new molecules
        id_mols.update(self.add_molecules(dict_mols)["data"])

        # Get molecules by index and translate back to dict
        tmp = self.get_molecules(list(id_mols.values()))
        id_mols_list = tmp["data"]
        meta["errors"].extend(tmp["meta"]["errors"])

        inv_id_mols = {v: k for k, v in id_mols.items()}

        for mol in id_mols_list:
            ret_mols[inv_id_mols[mol["id"]]] = mol

        meta["success"] = True
        meta["n_found"] = len(ret_mols)
        meta["missing"] = list(ordered_mol_dict.keys() - ret_mols.keys())

        # Rewind to flat last
        ret = []
        for ind in range(len(ordered_mol_dict)):
            if ind in ret_mols:
                ret.append(ret_mols[ind])
            else:
                ret.append(None)

        return {"meta": meta, "data": ret}

    def _add_generic(self, data, table, return_map=True):
        """
        Helper function that facilitates adding a record.
        """

        meta = {"errors": [], "n_inserted": 0, "success": False, "duplicates": [], "error_description": False}

        if len(data) == 0:
            ret = {}
            meta["success"] = True
            ret["meta"] = meta
            ret["data"] = {}
            return ret

        # Try/except for fully successful/partially unsuccessful adds
        error_skips = []
        try:
            tmp = self._tables[table].insert_many(data, ordered=False)
            meta["success"] = tmp.acknowledged
            meta["n_inserted"] = len(tmp.inserted_ids)
        except pymongo.errors.BulkWriteError as tmp:
            meta["success"] = False
            meta["n_inserted"] = tmp.details["nInserted"]
            for error in tmp.details["writeErrors"]:
                ukey = tuple(data[error["index"]][key] for key in self._table_indices[table])
                # Duplicate key errors, add to meta
                if error["code"] == 11000:
                    meta["duplicates"].append(ukey)
                else:
                    meta["errors"].append({"id": str(error["op"]["_id"]), "code": error["code"], "key": ukey})

                error_skips.append(error["index"])

            # Only duplicates, no true errors
            if len(meta["errors"]) == 0:
                meta["success"] = True
                meta["error_description"] = "Found duplicates"
            else:
                meta["error_description"] = "unknown"

        # Convert id in-place
        for d in data:
            d["id"] = str(d["_id"])
            del d["_id"]

        # Add id's of new keys
        skips = set(error_skips)
        rdata = []
        if return_map:
            for x in range(len(data)):
                if x in skips:
                    rdata.append(None)
                else:
                    rdata.append(data[x]["id"])

        ret = {"data": rdata, "meta": meta}

        return ret

    def _del_by_index(self, table, hashes, index="_id"):
        """
        Helper function that facilitates deletion based on hash.
        """

        if isinstance(hashes, str):
            hashes = [hashes]

        if index == "_id":
            _str_to_indices(hashes)

        return (self._tables[table].delete_many({index: {"$in": hashes}})).deleted_count

    def _get_generic(self, query, table, projection=None, allow_generic=False, limit=0):

        # TODO parse duplicates
        meta = storage_utils.get_metadata()

        data = []

        # Assume we want to lookup via unique key tuple
        if isinstance(query, (tuple, list)):
            keys = self._table_indices[table]
            len_key = len(keys)

            for q in query:
                if (len(q) == len_key) and isinstance(q, (list, tuple)):
                    q = {k: v for k, v in zip(keys, q)}
                else:
                    meta["errors"].append({"query": q, "error": "Malformed query"})
                    continue

                d = self._tables[table].find_one(q, projection=projection)
                if d is None:
                    meta["missing"].append(q)
                else:
                    data.append(d)

        elif isinstance(query, dict):

            # Handle specific ID query
            if "id" in query:
                ids, bad_ids = _str_to_indices_with_errors(query["id"])
                if bad_ids:
                    meta["errors"].append(("Bad Ids", bad_ids))

                query["_id"] = ids
                del query["id"]

            for k, v in query.items():
                if isinstance(v, (list, tuple)):
                    query[k] = {"$in": v}

            data = list(self._tables[table].find(query, projection=projection, limit=limit))
        else:
            meta["errors"] = "Malformed query"

        meta["n_found"] = len(data)
        if len(meta["errors"]) == 0:
            meta["success"] = True

        # Convert ID
        for d in data:
            d["id"] = str(d.pop("_id"))

        ret = {"meta": meta, "data": data}
        return ret

### Mongo molecule functions

    def add_molecules(self, data):
        """
        Adds molecules to the database.

        Parameters
        ----------
        data : dict of molecule-like JSON objects
            A {key: molecule} dictionary of molecules to input.

        Returns
        -------
        bool
            Whether the operation was successful.
        """

        # Build a dictionary of new molecules
        new_mols = {}
        for key, dmol in data.items():
            try:
                dmol = dmol.dict()
            except AttributeError:
                pass
            # All molecules must be fixed
            dmol["fix_com"] = True
            dmol["fix_orientation"] = True

            mol = interface.Molecule(**dmol, orient=False)
            new_mols[key] = mol

        new_kv_hash = {k: v.get_hash() for k, v in new_mols.items()}
        new_vk_hash = collections.defaultdict(list)
        for k, v in new_kv_hash.items():
            new_vk_hash[v].append(k)

        # We need to filter out what is already in the database
        old_mols = self.get_molecules(list(new_kv_hash.values()), index="hash")["data"]

        # If we have hash matches check to for duplicates
        key_mapper = {}
        for old_mol in old_mols:

            # This is the user provided key
            new_mol_keys = new_vk_hash[old_mol["identifiers"]["molecule_hash"]]
            new_mol = new_mols[new_mol_keys[0]]

            if new_mol.compare(old_mol):
                for x in new_mol_keys:
                    del new_mols[x]
                    key_mapper[x] = old_mol["id"]
            else:
                # If this happens, we need to think a bit about what to do
                # Effectively our molecule hash index now has duplicates.
                # This is *sort of* ok as we use uuid's for all internal projects.
                raise KeyError("!!! WARNING !!!: Hash collision detected")

        # Carefully make this flat
        new_hashes = set()
        new_inserts = []
        new_keys = []
        for new_key, new_mol in new_mols.items():
            data = new_mol.json_dict()
            data["identifiers"] = {}

            # Build new molecule hash
            data["molecule_hash"] = new_mol.get_hash()
            data["identifiers"]["molecule_hash"] = data["molecule_hash"]

            if data["molecule_hash"] in new_hashes:
                continue

            # Build chemical identifiers
            data["identifiers"]["molecular_formula"] = new_mol.get_molecular_formula()
            data["molecular_formula"] = data["identifiers"]["molecular_formula"]

            new_hashes |= set([data["molecule_hash"]])
            new_inserts.append(data)
            new_keys.append(new_key)

        ret = self._add_generic(new_inserts, "molecule", return_map=True)
        ret["meta"]["duplicates"].extend(list(key_mapper.keys()))
        ret["meta"]["validation_errors"] = []

        # If something went wrong, we cannot generate the full key map
        # Success should always be True as we are parsing duplicate above and *not* here.
        if ret["meta"]["success"] is False:
            ret["meta"]["error_description"] = "Major insert error."
            ret["data"] = key_mapper
            return ret

        # Add the new keys to the key map
        for mol in new_inserts:
            for x in new_vk_hash[mol["molecule_hash"]]:
                key_mapper[x] = mol["id"]

        ret["data"] = key_mapper

        return ret

    def get_molecules(self, molecule_ids, index="id"):

        ret = {"meta": storage_utils.get_metadata(), "data": []}

        try:
            index = storage_utils.translate_molecule_index(index)
        except KeyError as e:
            ret["meta"]["error_description"] = repr(e)
            return ret

        if not isinstance(molecule_ids, (list, tuple)):
            molecule_ids = [molecule_ids]

        bad_ids = []
        if index == "id":
            molecule_ids, bad_ids = _str_to_indices_with_errors(molecule_ids)

        # Don't include the hash or the molecular_formula in the returned result
        # Make the query
        query = {index + '__in': molecule_ids}
        data = Molecule.objects(**query).exclude("molecule_hash", "molecular_formula").as_pymongo()

        if data is None:
            data = []
        else:
            data = list(data)

        ret["meta"]["success"] = True
        ret["meta"]["n_found"] = len(data)
        if len(bad_ids):
            ret["meta"]["errors"].append(("Bad Ids", bad_ids))

        # Translate ID's back
        for r in data:
            r["id"] = str(r["_id"])
            del r["_id"]

        ret["data"] = data

        return ret

    def del_molecules(self, values, index="id"):
        """
        Removes a molecule from the database from its hash.

        Parameters
        ----------
        values : str or list of strs
            The hash of a molecule.

        Returns
        -------
        bool
            Number of deleted molecules.
        """

        index = storage_utils.translate_molecule_index(index)

        if isinstance(values, str):
            values = [values]

        query = {index + '__in': values}

        return Molecule.objects(**query).delete()

    def _doc_to_tuples(self, doc: db.Document, with_ids=True):
        """
        Todo: to be removed
        """
        if not doc:
            return

        table = doc._get_collection_name()

        d_json = json.loads(doc.to_json())
        d_json["id"] = str(doc.id)
        del d_json["_id"]
        ukey = tuple(str(doc[key]) for key in self._table_indices[table])
        if with_ids:
            rdata = (ukey, str(doc.id))
        else:
            rdata = ukey
        return rdata

    ### Mongo options functions

    def add_keywords(self, data: Union[Dict, List[Dict]]):
        """Add one option uniqely identified by 'program' and the 'name'.

        Parameters
        ----------
         data : dict or List[dict]
            The attribites of the 'option' or options to be inserted.
            Must include for each 'option':
                program : str, program name
                name : str, option name

        Returns
        -------
            A dict with keys: 'data' and 'meta'
            (see storage_utils.add_metadata())
            The 'data' part is a list of ids of the inserted options
            data['duplicates'] has the duplicate entries

        Notes
        ------
            Duplicates are not considered errors.

        """

        if not isinstance(data, Sequence):
            data = [data]

        meta = storage_utils.add_metadata()

        keywords = []
        try:
            for d in data:
                # search by index keywords not by all keys, much faster
                found = Keywords.objects(program=d['program'], hash_index=d['hash_index']).first()
                if not found:

                    # Make sure ID does not exists and is generated by Mongo
                    d = d.copy()
                    d.pop("id", None)

                    doc = Keywords(**d).save()
                    keywords.append(str(doc.id))
                    meta['n_inserted'] += 1
                else:
                    meta['duplicates'].append(d["hash_index"])  # TODO
                    keywords.append(str(found.id))
            meta["success"] = True
        except (mongoengine.errors.ValidationError, KeyError) as err:
            meta["validation_errors"].append(str(err))
            keywords.append(None)
        except Exception as err:
            meta['error_description'] = err
            keywords.append(None)

        ret = {"data": keywords, "meta": meta}
        return ret

    def get_keywords(self, id: str=None, program: str=None, hash_index: str=None, return_json: bool=True, with_ids: bool=True, limit=None):
        """Search for one (unique) option based on the 'program'
        and the 'name'. No overwrite allowed.

        Parameters
        ----------
        program : str
            program name
        name : str
            option name
        return_json : bool, optional
            Return the results as a json object
            Default is True
        with_ids : bool, optional
            Include the DB ids in the returned object (names 'id')
            Default is True
        limit : int, optional
            Maximum number of resaults to return.
            If this number is greater than the mongoengine_soket.max_limit then
            the max_limit will be returned instead.
            Default is to return the socket's max_limit (when limit=None or 0)

        Returns
        -------
            A dict with keys: 'data' and 'meta'
            (see storage_utils.get_metadata())
            The 'data' part is an object of the result or None if not found
        """

        meta = storage_utils.get_metadata()
        query = {}
        if program:
            query['program'] = program
        if hash_index:
            query['hash_index'] = hash_index
        if id:
            query['id'] = ObjectId(id)
        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            data = Keywords.objects(**query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            rdata = [d.to_json_obj(with_ids) for d in data]
        else:
            rdata = data

        return {"data": rdata, "meta": meta}

    def get_add_keywords_mixed(self, data):
        """
        Get or add the given options (if they don't exit).
        Keywords are given in a mixed format, either as a dict of mol data
        or as existing mol id

        TODO: to be split into get by_id and get_by_data
        """

        meta = storage_utils.get_metadata()

        ids = []
        for idx, kw in enumerate(data):
            if isinstance(kw, str):
                ids.append(kw)

            # New dictionary construct and add
            elif isinstance(kw, dict):

                kw = interface.models.common_models.KeywordSet(**kw)

                new_id = self.add_keywords([kw.json_dict()])["data"][0]
                ids.append(new_id)

            elif isinstance(kw, interface.models.common_models.KeywordSet):
                new_id = self.add_keywords([kw.json_dict()])["data"][0]
                ids.append(new_id)
            else:
                meta["errors"].append((idx, "Data type not understood"))
                ids.append(None)

        missing = []
        ret = []
        for idx, id in enumerate(ids):
            if id is None:
                ret.append(None)
                missing.append(idx)
                continue

            try:
                ret.append(self.get_keywords(id=id)["data"][0])
            except bson.errors.InvalidId:
                ret.append(None)
                missing.append(idx)

        meta["success"] = True
        meta["n_found"] = len(ret) - len(missing)
        meta["missing"] = missing

        return {"meta": meta, "data": ret}

    def del_keywords(self, id):
        """
        Removes a option set from the database based on its keys.

        Parameters
        ----------
        program : str
            The program of the option set
        name : str
            The name of the option set

        Returns
        -------
        int
           number of deleted documents
        """

        # monogoengine
        count = 0
        kw = Keywords.objects(id=ObjectId(id))
        if kw:
            count = kw.delete()

        return count

    ### Mongo database functions

    # def add_collection(self, data, overwrite=False):
    def add_collection(self, collection: str, name: str, data, overwrite: bool=False):
        """Add (or update) a collection to the database.

        Parameters
        ----------
        collection : str
        name : str
        data : dict
        overwrite : bool
            Update existing collection

        Returns
        -------
        A dict with keys: 'data' and 'meta'
            (see storage_utils.add_metadata())
            The 'data' part is the id of the inserted document or none

        Notes
        -----
        ** Change: The data doesn't have to include the ID, the document
        is identified by the (collection, name) pairs.
        ** Change: New fields will be added to the collection, but existing won't
            be removed.
        """

        meta = storage_utils.add_metadata()
        col_id = None
        try:

            if ("id" in data) and (data["id"] == "local"):
                data.pop("id", None)

            if overwrite:
                # may use upsert=True to add or update
                col = Collection.objects(collection=collection, name=name).update_one(**data)
            else:
                col = Collection(collection=collection, name=name, **data).save()

            meta['success'] = True
            meta['n_inserted'] = 1
            col_id = str(col.id)
        except Exception as err:
            meta['error_description'] = str(err)

        ret = {'data': col_id, 'meta': meta}
        return ret

    # def get_collections(self, keys, projection=None):
    def get_collections(self,
                        collection: str=None,
                        name: str=None,
                        return_json: bool=True,
                        with_ids: bool=True,
                        limit: int=None):
        """Get collection by collection and/or name

        Parameters
        ----------
        collection : str, optional
        name : str, optional
        return_json : bool
        with_ids : bool
        limit : int

        Returns
        -------
        A dict with keys: 'data' and 'meta'
            The data is a list of the collections found
        """

        meta = storage_utils.get_metadata()
        query = {}
        if collection:
            query['collection'] = collection
        if name:
            query['name'] = name
        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            data = Collection.objects(**query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            rdata = [d.to_json_obj(with_ids) for d in data]
        else:
            rdata = data

        return {"data": rdata, "meta": meta}

    def del_collection(self, collection: str, name: str):
        """
        Remove a collection from the database from its keys.

        Parameters
        ----------
        collection: str
            Collection type
        name : str
            Collection name

        Returns
        -------
        int
            Number of documents deleted
        """

        return Collection.objects(collection=collection, name=name).delete()

    # -------------------------- Results functions ----------------------------
    #
    # def add_result(
    #         self,
    #         program: str,
    #         method: str,
    #         driver: str,
    #         molecule: str,  # Molecule id
    #         basis: str,
    #         options: str,
    #         data: dict,
    #         return_json=True,
    #         with_ids=True):
    #     """ Add one result
    #     """

    def add_results(self, data: List[dict], update_existing: bool=False, return_json=True):
        """
        Add results from a given dict. The dict should have all the required
        keys of a result.

        Parameters
        ----------
        data : list of dict
            Each dict must have:
            program, driver, method, basis, options, molecule
            Where molecule is the molecule id in the DB
            In addition, it should have the other attributes that it needs
            to store
        update_existing : bool (default False)
            Update existing results

        Returns
        -------
            Dict with keys: data, meta
            Data is the ids of the inserted/updated/existing docs
        """

        for d in data:
            for i in self._lower_results_index:
                if d[i] is None:
                    continue

                d[i] = d[i].lower()

        meta = storage_utils.add_metadata()

        results = []
        # try:
        for d in data:

            d.pop("id", None)
            # search by index keywords not by all keys, much faster
            doc = Result.objects(
                program=d['program'],
                name=d['driver'],
                method=d['method'],
                basis=d['basis'],
                keywords=d['keywords'],
                molecule=d['molecule'])

            if doc.count() == 0 or update_existing:
                if not isinstance(d['molecule'], ObjectId):
                    d['molecule'] = ObjectId(d['molecule'])
                doc = doc.upsert_one(**d)
                results.append(str(doc.id))
                meta['n_inserted'] += 1
            else:
                id = str(doc.first().id)
                meta['duplicates'].append(id)  # TODO
                # If new or duplicate, add the id to the return list
                results.append(id)
        meta["success"] = True
        # except (mongoengine.errors.ValidationError, KeyError) as err:
        #     meta["validation_errors"].append(err)
        # except Exception as err:
        #     meta['error_description'] = err

        ret = {"data": results, "meta": meta}
        return ret

    def get_results_by_id(self, id: List[str]=None, projection=None, return_json=True, with_ids=True):
        """
        Get list of Results using the given list of Ids

        Parameters
        ----------
        id : List of str
            Ids of the results in the DB
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        return_json : bool, default is True
            Return the results as a list of json inseated of objects
        with_ids: bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()

        data = []
        # try:
        if projection:
            data = Result.objects(id__in=id).only(*projection).limit(self._max_limit)
        else:
            data = Result.objects(id__in=id).limit(self._max_limit)

        meta["n_found"] = data.count()
        meta["success"] = True
        # except Exception as err:
        #     meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def get_results_count(self):
        """
        TODO: just return the count, used for big queries

        Returns
        -------

        """
        pass

    def get_results(self,
                    program: str=None,
                    method: str=None,
                    basis: str=None,
                    molecule: str=None,
                    driver: str=None,
                    keywords: str=None,
                    status: str='COMPLETE',
                    projection=None,
                    limit: int=None,
                    skip: int=None,
                    return_json=True,
                    with_ids=True):
        """

        Parameters
        ----------
        program : str
        method : str
        basis : str
        molecule : str
            Molecule id in the DB
        driver : str
        keywords : str
            The id of the option in the DB
        status : bool, default is 'COMPLETE'
            The status of the result: 'COMPLETE', 'INCOMPLETE', or 'ERROR'
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is None TODO
            skip the first 'skip' resaults. Used to paginate
        return_json : bool, deafult is True
            Return the results as a list of json inseated of objects
        with_ids : bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()
        query = {}
        parsed_query = {}
        if program:
            query['program'] = program
        if method:
            query['method'] = method
        if basis:
            query['basis'] = basis
        if molecule:
            query['molecule'], _ = _str_to_indices_with_errors(molecule)
        if driver:
            query['driver'] = driver
        if keywords:
            query['keywords'] = keywords
        if status:
            query['status'] = status

        for key, value in query.items():
            if key == "molecule":
                parsed_query[key + "__in"] = query[key]
            elif key == "status":
                parsed_query[key] = value
            elif isinstance(value, (list, tuple)):
                parsed_query[key + "__in"] = [v.lower() for v in value]
            else:
                parsed_query[key] = value.lower()

        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            if projection:
                data = Result.objects(**parsed_query).only(*projection).limit(q_limit)
            else:
                data = Result.objects(**parsed_query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def get_results_by_task_id(self,
                               task_id: Union[List[str], str],
                               projection=None,
                               limit: int=None,
                               return_json=True):
        """

        Parameters
        ----------
        task_id : List of str or str
            Task id that ran the results
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        return_json : bool, deafult is True
            Return the results as a list of json instead of objects

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()
        query = {}

        if isinstance(task_id, (list, tuple)):
            query['task_id__in'] = task_id
        else:
            query['task_id'] = task_id

        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            if projection:
                data = Result.objects(**query).only(*projection).limit(q_limit)
            else:
                data = Result.objects(**query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj() for d in data]

        return {"data": data, "meta": meta}

    def del_results(self, ids: List[str]):
        """
        Removes results from the database using their ids
        (Should be cautious! other tables maybe referencing results)

        Parameters
        ----------
        ids : list of str
            The Ids of the results to be deleted

        Returns
        -------
        int
            number of results deleted
        """

        obj_ids = [ObjectId(x) for x in ids]

        return Result.objects(id__in=obj_ids).delete()

### Mongo procedure/service functions

    def add_procedures(self, data: List[dict], update_existing: bool=False, return_json=True):
        """
        Add procedures from a given dict. The dict should have all the required
        keys of a result.

        Parameters
        ----------
        data : list of dict
            Each dict must have:
            procedure, program, keywords, qc_meta, hash_index
            In addition, it should have the other attributes that it needs
            to store
        update_existing : bool (default False)
            Update existing results

        Returns
        -------
            Dict with keys: data, meta
            Data is the ids of the inserted/updated/existing docs
        """

        meta = storage_utils.add_metadata()

        results = []
        # try:
        for d in data:
            # search by hash index
            d.pop("id", None)
            doc = Procedure.objects(hash_index=d['hash_index'])

            if doc.count() == 0 or update_existing:
                doc = doc.upsert_one(**d)
                results.append(str(doc.id))
                meta['n_inserted'] += 1
            else:
                id = str(doc.first().id)
                meta['duplicates'].append(id)  # TODO
                # If new or duplicate, add the id to the return list
                results.append(id)
        meta["success"] = True
        # except (mongoengine.errors.ValidationError, KeyError) as err:
        #     meta["validation_errors"].append(err)
        # except Exception as err:
        #     meta['error_description'] = err

        ret = {"data": results, "meta": meta}
        return ret

    def get_procedures(self,
                       procedure: str=None,
                       program: str=None,
                       hash_index: str=None,
                       ids: List[str]=None,
                       status: str='COMPLETE',
                       projection=None,
                       limit: int=None,
                       skip: int=None,
                       return_json=True,
                       with_ids=True):
        """

        Parameters
        ----------
        procedure : str
        program : str
        hash_index : str
        ids : str
        status : bool, default is 'COMPLETE'
            The status of the result: 'COMPLETE', 'INCOMPLETE', or 'ERROR'
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is None TODO
            skip the first 'skip' resaults. Used to paginate
        return_json : bool, deafult is True
            Return the results as a list of json inseated of objects
        with_ids : bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()
        query = {}
        parsed_query = {}
        if procedure:
            query['procedure'] = procedure
        if program:
            query['program'] = program
        if hash_index:
            query['hash_index'] = hash_index
        if ids:
            query['ids'] = ids
        if status:
            query['status'] = status

        for key, value in query.items():
            if key == "status":
                parsed_query[key] = value
            elif isinstance(value, (list, tuple)):
                parsed_query[key + "__in"] = [v.lower() for v in value]
            else:
                parsed_query[key] = value.lower()

        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            if projection:
                data = Procedure.objects(**parsed_query).only(*projection).limit(q_limit)
            else:
                data = Procedure.objects(**parsed_query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def get_procedures_by_id(self,
                             id: List[str]=None,
                             hash_index: List[str]=None,
                             projection=None,
                             return_json=True,
                             with_ids=True):
        """
        Get list of Procedures using the given list of Ids

        Parameters
        ----------
        id : List of str
            Ids of the results in the DB
        hash_index: List or str
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        return_json : bool, default is True
            Return the results as a list of json instead of objects
        with_ids: bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()

        query, parsed_query = {}, {}
        if id:
            query['id'] = id
        if hash_index:
            query['hash_index'] = hash_index

        for key, value in query.items():
            if isinstance(value, (list, tuple)):
                parsed_query[key + "__in"] = value
            else:
                parsed_query[key] = value

        data = []
        # try:
        if projection:
            data = Procedure.objects(**parsed_query).only(*projection).limit(self._max_limit)
        else:
            data = Procedure.objects(**parsed_query).limit(self._max_limit)

        meta["n_found"] = data.count()
        meta["success"] = True
        # except Exception as err:
        #     meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def get_procedures_by_task_id(self,
                                  task_id: Union[List[str], str],
                                  projection=None,
                                  limit: int=None,
                                  return_json=True):
        """

        Parameters
        ----------
        task_id : List of str or str
            Task id that ran the procedure
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        return_json : bool, deafult is True
            Return the results as a list of json instead of objects

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()
        query = {}

        if isinstance(task_id, (list, tuple)):
            query['task_id__in'] = task_id
        else:
            query['task_id'] = task_id

        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            if projection:
                data = Procedure.objects(**query).only(*projection).limit(q_limit)
            else:
                data = Procedure.objects(**query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            data = [d.to_json_obj() for d in data]

        return {"data": data, "meta": meta}

    def update_procedure(self, hash_index, data):
        """
        TODO: to be updated with needed
        """

        if 'id' in data and data['id'] is None:
            data.pop("id", None)

        update = {}
        not_allowed_keys = []
        # create safe query with allowed keys only
        # shouldn't be allowed to update status manually
        for key, value in data.items():
            if key not in ['procedure', 'program', 'status', 'task_id']:  # FIXME: what else?
                update[key] = value
            else:
                not_allowed_keys.append(key)

        if not_allowed_keys:
            logging.warning('Trying to update Procedure immutable keywords ' +
                            '"{}", skipping'.format(not_allowed_keys))

        modified_count = Procedure.objects(hash_index=hash_index).update(**update, modified_on=dt.utcnow())

        return modified_count

    def add_services(self, data: List[dict], update_existing: bool=False, return_json=True):
        """
        Add services from a given dict.

        Parameters
        ----------
        data : list of dict
        update_existing: bool, default False

        Returns
        -------
            Dict with keys: data, meta
            Data is the hash_index of the inserted/existing docs
        """

        meta = storage_utils.add_metadata()

        services = []
        # try:
        for d in data:
            # search by hash index
            d.pop("id", None)
            doc = ServiceQueue.objects(hash_index=d['hash_index'])

            if doc.count() == 0 or update_existing:
                doc = doc.upsert_one(**d)
                services.append(doc.hash_index)
                meta['n_inserted'] += 1
            else:
                # id = str(doc.first().id)
                # By D2: Right now services expect hash return
                # This and bad and should be fixed
                hash_index = doc.first().hash_index
                meta['duplicates'].append(hash_index)
                # If new or duplicate, add the to the return list
                services.append(hash_index)
        meta["success"] = True
        # except (mongoengine.errors.ValidationError, KeyError) as err:
        #     meta["validation_errors"].append(err)
        # except Exception as err:
        #     meta['error_description'] = err

        ret = {"data": services, "meta": meta}
        return ret

    def get_services(self,
                     id: Union[List[str], str]=None,
                     hash_index: Union[List[str], str]=None,
                     status: str=None,
                     projection=None,
                     limit: int=None,
                     return_json=True):
        """

        Parameters
        ----------
        id / hash_index : List of str or str
            service id / hash_index that ran the results
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        return_json : bool, deafult is True
            Return the results as a list of json instead of objects

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()
        query = {}

        if isinstance(id, (list, tuple)):
            query['id__in'] = id
        elif id:
            query['id'] = id

        if isinstance(hash_index, (list, tuple)):
            query['hash_index__in'] = hash_index
        elif hash_index:
            query['hash_index'] = hash_index

        if status:
            query['status'] = status

        q_limit = int(limit if limit and limit < self._max_limit else self._max_limit)

        data = []
        # try:
        if projection:
            services = ServiceQueue.objects(**query).only(*projection).limit(q_limit)
        else:
            services = ServiceQueue.objects(**query).limit(q_limit)

        meta["n_found"] = services.count()
        meta["success"] = True
        # except Exception as err:
        #     meta['error_description'] = str(err)

        if services.count() and return_json:
            data = [d.to_json_obj() for d in services]

        return {"data": data, "meta": meta}


    def update_services(self, updates):

        match_count = 0
        modified_count = 0
        for uid, data in updates:
            if 'id' in data and data['id'] is None:
                data.pop("id", None)
            result = self._tables["service_queue"].replace_one({"_id": ObjectId(uid)}, data)
            match_count += result.matched_count
            modified_count += result.modified_count
        return (match_count, modified_count)

    def del_services(self, values, index="id"):

        index = _translate_id_index(index)

        return self._del_by_index("service_queue", values, index=index)

### Mongo queue handling functions

    def queue_submit(self, data: List[Dict]):
        """Submit a list of tasks to the queue.
        Tasks are unique by their base_result, which should be inserted into
        the DB first before submitting it's corresponding task to the queue
        (with result.status='INCOMPLETE' as the default)
        The default task.status is 'WAITING'

        Duplicate tasks sould be a rare case.
        Hooks are merged if the task already exists

        Parameters
        ----------
        data : list of tasks (dict)
            A task is a dict, with the following fields:
            - hash_index: idx, not used anymore
            - spec: dynamic field (dict-like), can have any structure
            - hooks: list of any objects representing listeners (for now)
            - tag: str
            - base_results: tuple (required), first value is the class type
             of the result, {'results' or 'procedure'). The second value is
             the ID of the result in the DB. Example:
             "base_result": ('results', result_id)

        Returns
        -------
        dict (data and meta)
            'data' is a list of the IDs of the tasks IN ORDER, including
            duplicates. An errored task has 'None' in its ID
            meta['duplicates'] has the duplicate tasks
        """

        meta = storage_utils.add_metadata()

        results = []
        for d in data:
            try:
                if not isinstance(d['base_result'], tuple):
                    raise Exception("base_result must be a tuple not {}.".format(type(d['base_result'])))

                # If saved as DBRef, then use raw query to retrieve (avoid this)
                # if d['base_result'][0] in ('results', 'procedure'):
                #     base_result = DBRef(d['base_result'][0], d['base_result'][1])

                d.pop("id", None)
                result_obj = None
                if d['base_result'][0] == 'results':
                    result_obj = Result(id=d['base_result'][1])
                elif d['base_result'][0] == 'procedure':
                    result_obj = Procedure(id=d['base_result'][1])
                else:
                    raise TypeError("Base_result type must be 'results' or 'procedure',"
                                    " {} is given.".format(d['base_result'][0]))
                task = TaskQueue(**d)
                task.base_result = result_obj
                task.save()
                result_obj.update(task_id=str(task.id))  # update bidirectional rel
                results.append(str(task.id))
                meta['n_inserted'] += 1
            except mongoengine.errors.NotUniqueError as err:  # rare case
                # If results is stored as DBRef, get it with:
                # task = TaskQueue.objects(__raw__={'base_result': base_result}).first()  # avoid

                # If base_result is stored as a Result or Procedure class, get it with:
                task = TaskQueue.objects(base_result=result_obj).first()
                self.logger.warning('queue_submit got a duplicate task: ', task.to_mongo())
                if d['hooks']:  # merge hooks
                    task.hooks.extend(d['hooks'])
                    task.save()
                results.append(str(task.id))
                meta['duplicates'].append(self._doc_to_tuples(task, with_ids=False))  # TODO
            except Exception as err:
                meta["success"] = False
                meta["errors"].append(str(err))
                results.append(None)

        meta["success"] = True

        ret = {"data": results, "meta": meta}
        return ret

    def queue_get_next(self, manager, limit=100, tag=None, as_json=True):
        """TODO: needs to be done a transcation"""

        # Figure out query, tagless has no requirements
        query = {"status": "WAITING"}
        if tag is not None:
            query["tag"] = tag

        found = TaskQueue.objects(**query).limit(limit).order_by('created_on')

        query = {"_id": {"$in": [x.id for x in found]}}

        # update_many using pymongo in one DB access
        upd = TaskQueue._get_collection().update_many(
            query, {"$set": {
                "status": "RUNNING",
                "modified_on": dt.utcnow(),
                "manager": manager,
            }})

        if as_json:
            found = [task.to_json_obj() for task in found]

        if upd.modified_count != len(found):
            self.logger.warning("QUEUE: Number of found projects does not match the number of updated projects.")

        return found

    # def get_queue_(self, query, projection=None):
    #     """TODO: to be replaced with a specific query, add limit"""
    #
    #     return self._get_generic(query, "task_queue", allow_generic=True, projection=projection)

    def get_queue(self,
                  id=None,
                  hash_index=None,
                  program=None,
                  status: str=None,
                  projection=None,
                  limit: int=None,
                  skip: int=None,
                  return_json=True,
                  with_ids=True):
        """
        TODO: check what query keys are needs
        Parameters
        ----------
        id : list or str
            Id of the task
        Hash_index
        status : bool, default is None (find all)
            The status of the task: 'COMPLETE', 'RUNNING', 'WAITING', or 'ERROR'
        projection : list/set/tuple of keys, default is None
            The fields to return, default to return all
        limit : int, default is None
            maximum number of results to return
            if 'limit' is greater than the global setting self._max_limit,
            the self._max_limit will be returned instead
            (This is to avoid overloading the server)
        skip : int, default is None TODO
            skip the first 'skip' resaults. Used to paginate
        return_json : bool, deafult is True
            Return the results as a list of json inseated of objects
        with_ids : bool, default is True
            Include the ids in the returned objects/dicts

        Returns
        -------
        Dict with keys: data, meta
            Data is the objects found
        """

        meta = storage_utils.get_metadata()
        query = {}
        parsed_query = {}
        if program:
            query['program'] = program
        if id is not None:
            query['id'] = id
        if hash_index:
            query['hash_index'] = hash_index
        if status:
            query['status'] = status

        for key, value in query.items():
            if isinstance(value, (list, tuple)):
                parsed_query[key + "__in"] = value
            else:
                parsed_query[key] = value

        q_limit = limit if limit and limit < self._max_limit else self._max_limit

        data = []
        try:
            if projection:
                data = TaskQueue.objects(**parsed_query).only(*projection).limit(q_limit)
            else:
                data = TaskQueue.objects(**parsed_query).limit(q_limit)

            meta["n_found"] = data.count()
            meta["success"] = True
        except Exception as err:
            meta['error_description'] = str(err)

        if return_json:
            if return_json:
                data = [d.to_json_obj(with_ids) for d in data]

        return {"data": data, "meta": meta}

    def queue_get_by_id(self, ids: List[str], limit: int=100, as_json: bool=True):
        """Get tasks by their IDs

        Parameters
        ----------
        ids : list of str
            List of the task Ids in the DB
        limit : int (optional)
            max number of returned tasks. If limit > max_limit, max_limit
            will be returned instead (safe query)
        as_json : bool
            Return tasks as JSON

        Returns
        -------
        list of the found tasks
        """

        q_limit = limit if limit and limit < self._max_limit else self._max_limit
        found = TaskQueue.objects(id__in=ids).limit(q_limit)

        if as_json:
            found = [task.to_json_obj() for task in found]

        return found

    def queue_mark_complete(self, task_ids: List[str]) -> int:
        """Update the given tasks as complete
        Note that each task is already pointing to its result location
        Mark the corresponding result/procedure as complete

        Parameters
        ----------
        task_ids : list
            IDs of the tasks to mark as COMPLETE

        Returns
        -------
        int
            Updated count
        """

        # If using replica sets, we can use as a transcation:
        # with self.client.start_session() as session:
        #     with session.start_transaction():
        #         # automatically calls ClientSession.commit_transaction().
        #         # If the block exits with an exception, the transaction
        #         # automatically calls ClientSession.abort_transaction().
        #           found = TaskQueue._get_collection().update_many(
        #                               {'id': {'$in': task_ids}},
        #                               {'$set': {'status': 'COMPLETE'}},
        #                               session=session)
        #         # next, Update results

        tasks = TaskQueue.objects(id__in=task_ids).update(status='COMPLETE', modified_on=dt.utcnow())
        results = Result.objects(task_id__in=task_ids).update(status='COMPLETE', modified_on=dt.utcnow())
        procedures = Procedure.objects(task_id__in=task_ids).update(status='COMPLETE', modified_on=dt.utcnow())

        # This should not happen unless there is data inconsistency in the DB
        if results + procedures < tasks:
            logging.error("Some tasks don't reference results or procedures correctly!"
                          "Tasks: {}, Results: {}, procedures: {}. ".format(tasks, results, procedures))
        return tasks

    def queue_mark_error(self, data):
        """update the given tasks as errored
        Mark the corresponding result/procedure as complete

        """

        bulk_commands = []
        task_ids = []
        for task_id, msg in data:
            update = {
                "$set": {
                    "status": "ERROR",
                    "error": msg,
                    "modified_on": dt.utcnow(),
                }
            }
            bulk_commands.append(pymongo.UpdateOne({"_id": ObjectId(task_id)}, update))
            task_ids.append(task_id)

        if len(bulk_commands) == 0:
            return

        ret = TaskQueue._get_collection().bulk_write(bulk_commands, ordered=False).modified_count
        Result.objects(task_id__in=task_ids).update(status='ERROR', modified_on=dt.utcnow())
        Procedure.objects(task_id__in=task_ids).update(status='ERROR', modified_on=dt.utcnow())

        return ret

    def queue_reset_status(self, manager: str, reset_running: bool=True, reset_error: bool=False) -> int:
        """
        Reset the status of the tasks that a manager owns from Running to Waiting
        If reset_error is True, then also reset errored tasks AND its results/proc

        Parameters
        ----------
        manager : str
            The manager name to reset the status of
        reset_running : str (optional), default is True
            If True, reset running tasks to be waiting
        reset_error : str (optional), default is False
            If True, also reset errored tasks to be waiting,
            also update results/proc to be INCOMPLETE

        Returns
        -------
        int
            Updated count
        """

        if not (reset_running or reset_error):
            # nothing to do
            return 0

        # Update results and procedures if reset_error
        if reset_error:
            task_ids = TaskQueue.objects(manager=manager, status="ERROR").only('id')
            Result.objects(task_id__in=task_ids).update(status='INCOMPLETE', modified_on=dt.utcnow())
            Procedure.objects(task_id__in=task_ids).update(status='INCOMPLETE', modified_on=dt.utcnow())

        status = []
        if reset_running:
            status.append("RUNNING")
        if reset_error:
            status.append("ERROR")

        updated = TaskQueue.objects(
            manager=manager, status__in=status).update(
                status="WAITING", modified_on=dt.utcnow())

        return updated

    def handle_hooks(self, hooks):

        # Very dangerous, we need to modify this substatially
        # Does not currently handle multiple identical commands
        # Only handles service updates

        bulk_commands = []
        for hook_list in hooks:
            for hook in hook_list:
                commands = {}
                for com in hook["updates"]:
                    commands["$" + com[0]] = {com[1]: com[2]}

                upd = pymongo.UpdateOne({"_id": ObjectId(hook["document"][1])}, commands)
                bulk_commands.append(upd)

        if len(bulk_commands) == 0:
            return

        ret = self._tables["service_queue"].bulk_write(bulk_commands, ordered=False)
        return ret

### QueueManagers

    def manager_update(self, name, **kwargs):

        inc_count = {
            # Increment relevant data
            "inc__submitted": kwargs.pop("submitted", 0),
            "inc__completed": kwargs.pop("completed", 0),
            "inc__returned": kwargs.pop("returned", 0),
            "inc__failures": kwargs.pop("failures", 0)
        }

        upd = {key: kwargs[key] for key in QueueManager._fields_ordered if key in kwargs}

        QueueManager.objects()  # init
        manager = QueueManager.objects(name=name)
        if manager:  # existing
            upd.update(inc_count)
            num_updated = manager.update(**upd, modified_on=dt.utcnow())
        else:  # create new, ensures defaults and validations
            QueueManager(name=name, **upd).save()
            num_updated = 1

        return num_updated == 1

    def get_managers(self, name: str=None, status: str=None, modified_before=None):

        query = {}
        if name:
            query["name"] = name
        if modified_before:
            query["modified_on__lt"] = modified_before
        if status:
            query["status"] = status

        data = QueueManager.objects(**query)

        meta = storage_utils.get_metadata()
        meta["success"] = True
        meta["n_found"] = data.count()

        data = [x.to_json_obj(with_id=False) for x in data]

        return {"data": data, "meta": meta}

### Users

    def add_user(self, username, password, permissions=["read"]):
        """
        Adds a new user and associated permissions.

        Passwords are stored using bcrypt.

        Parameters
        ----------
        username : str
            New user's username
        password : str
            The user's password
        permissions : list of str, optional
            The associated permissions of a user ['read', 'write', 'compute', 'queue', 'admin']

        Returns
        -------
        tuple
            Successful insert or not
        """

        hashed = bcrypt.hashpw(password.encode("UTF-8"), bcrypt.gensalt(6))
        try:
            User(username=username, password=hashed, permissions=permissions).save()
            return True
        except mongoengine.errors.NotUniqueError:
            return False

    def verify_user(self, username, password, permission):
        """
        Verifies if a user has the requested permissions or not.

        Passwords are store and verified using bcrypt.

        Parameters
        ----------
        username : str
            The username to verify
        password : str
            The password associated with the username
        permission : str
            The associated permissions of a user ['read', 'write', 'compute', 'queue', 'admin']

        Returns
        -------
        tuple
            A tuple of (success flag, failure string)

        Examples
        --------

        >>> db.add_user("george", "shortpw")

        >>> db.verify_user("george", "shortpw", "read")
        True

        >>> db.verify_user("george", "shortpw", "admin")
        False

        """

        if self._bypass_security:
            return (True, "Success")

        data = User.objects(username=username).first()
        if data is None:
            return (False, "User not found.")

        pwcheck = bcrypt.checkpw(password.encode("UTF-8"), data.password)
        if pwcheck is False:
            return (False, "Incorrect password.")

        # Admin has access to everything
        if (permission.lower() not in data.permissions) and ("admin" not in data.permissions):
            return (False, "User has insufficient permissions.")

        return (True, "Success")

    def remove_user(self, username):
        """Removes a user from the MongoDB Tables

        Parameters
        ----------
        username : str
            The username to remove

        Returns
        -------
        bool
            If the operation was successful or not.
        """
        return User.objects(username=username).delete() == 1



"""
A set of utility functions to help the Collections
"""

import math

__registered_collections = {}


def nCr(n, r):
    """
    Compute the binomial coefficient n! / (k! * (n-k)!)

    Parameters
    ----------
    n : int
        Number of samples
    r : int
        Denominator

    Returns
    -------
    ret : int
        Value
    """
    return math.factorial(n) / math.factorial(r) / math.factorial(n - r)


def register_collection(collection):
    """Registers a collection for the factory

    Parameters
    ----------
    collection : Collection
        The Collection class to be registered
    """

    class_name = collection.__name__.lower()
    if class_name in __registered_collections:
        raise KeyError("Collection type '{}' already registered".format(class_name))
    __registered_collections[class_name] = collection


def collection_factory(data, client=None):
    """Creates a new Collection class from a JSON blob.

    Parameters
    ----------
    data : dict
        The JSON blob to create a new class from.
    client : client.FractalClient
        A Portal client to connected to a server

    Returns
    -------
    Collection
        A ODM of the data.
    """
    if "collection" not in data:
        raise KeyError("Attempted to create Collection from JSON, but no `collection` field found.")

    if data["collection"].lower() not in __registered_collections:
        raise KeyError("Attempted to create Collection of unknown type '{}'.".format(data["collection"]))

    return __registered_collections[data["collection"].lower()].from_json(data, client=client)



"""
Tests the on-node procedures compute capabilities.
"""

import pytest
import requests

import qcfractal
import qcfractal.interface as portal
from qcfractal import testing

_users = {
    "read": {
        "pw": "hello",
        "perm": ["read"]
    },
    "write": {
        "pw": "something",
        "perm": ["read", "write"]
    },
    "admin": {
        "pw": "something",
        "perm": ["read", "write", "compute", "admin"]
    }
}


@pytest.fixture(scope="module")
def sec_server(request):
    """
    Builds a server instance with the event loop running in a thread.
    """

    # Check mongo
    testing.check_active_mongo_server()

    storage_name = "qcf_local_server_auth_test"

    with testing.loop_in_thread() as loop:

        # Build server, manually handle IOLoop (no start/stop needed)
        server = qcfractal.FractalServer(
            port=testing.find_open_port(), storage_project_name=storage_name, loop=loop, security="local")

        # Clean and re-init the databse
        server.storage.client.drop_database(server.storage._project_name)

        # Add local users
        for k, v in _users.items():
            assert server.storage.add_user(k, _users[k]["pw"], _users[k]["perm"])

        yield server


### Tests the compute queue stack
def test_security_auth_decline_none(sec_server):
    with pytest.raises(requests.exceptions.HTTPError) as excinfo:
        client = portal.FractalClient(sec_server)

    assert "user not found" in str(excinfo.value).lower()


def test_security_auth_bad_ssl(sec_server):
    with pytest.raises(requests.exceptions.SSLError) as excinfo:
        client = portal.FractalClient.from_file({
            "address": sec_server.get_address(),
            "username": "read",
            "password": _users["write"]["pw"],
            "verify": True
        })

    assert "ssl handshake" in str(excinfo.value).lower()
    assert "verify=false" in str(excinfo.value).lower()


def test_security_auth_decline_bad_user(sec_server):
    with pytest.raises(requests.exceptions.HTTPError) as excinfo:
        client = portal.FractalClient.from_file({
            "address": sec_server.get_address(),
            "username": "hello",
            "password": "something",
            "verify": False
        })

    assert "user not found" in str(excinfo.value).lower()


def test_security_auth_accept(sec_server):

    client = portal.FractalClient(sec_server, username="write", password=_users["write"]["pw"])

    r = client.add_molecules({})
    r = client.get_molecules([])



"""
The schema for various database specifications
"""

database_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "properties": {
        "program": {
            "type": "string",
            "description": "The QC program to use"
        },
        "options": {
            "type": "string",
            "description": "The name of the options set to use."
        },
        "model": {
            "type": "object",
            "description": "The physical model to apply to the system",
            "properties": {
                "method": {
                    "type": "string",
                },
                "basis": {
                    "type": "string"
                }
            }
        },
        "molecule_hashes": {
            "type": "array",
            "items": {
                "type": "string",
                "description": "The required compute hashes."
            }
        }
    }
}



import setuptools
import versioneer

if __name__ == "__main__":
    setuptools.setup(
        name='qcfractal',
        description='A high throughput computing and database tool for quantum chemistry.',
        author='Daniel Smith',
        author_email='dgasmith@vt.edu',
        url="https://github.com/molssi/qcfractal",
        license='BSD-3C',
        include_package_data=True,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        packages=setuptools.find_packages(),
        install_requires=[
            # Base requires
            'bcrypt',
            'cryptography',
            'numpy>=1.7',
            'pandas',
            'pydantic>=0.19',
            'pymongo>=3.0',
            'requests',
            'tornado',

            # Database
            'mongoengine',

            # QCArchive depends
            'qcengine>=0.5.1',
            'qcelemental>=0.2.4',

            # Testing
            'pytest',
        ],
        entry_points={
            "console_scripts": [
                "qcfractal-server=qcfractal.cli.qcfractal_server:main",
                "qcfractal-manager=qcfractal.cli.qcfractal_manager:main",
            ]
        },
        extras_require={
            'docs': [
                'sphinx==1.2.3',  # autodoc was broken in 1.3.1
                'sphinxcontrib-napoleon',
                'sphinx_rtd_theme',
                'numpydoc',
            ],
            'tests': [
                'pytest',
                'pytest-cov',
            ],
        },
        tests_require=[
            'pytest',
            'pytest-cov',
        ],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3',
        ],
        zip_safe=True,
    )



"""
Contains a number of utility functions for storage sockets
"""

import json

# Constants
_get_metadata = json.dumps({"errors": [], "n_found": 0, "success": False, "missing": [],
                            "error_description": False})

_add_metadata = json.dumps({"errors": [], "n_inserted": 0, "success": False, "duplicates": [],
                            "error_description": False, "validation_errors": []})


def translate_molecule_index(index):
    if index in ["_id", "ids"]:
        return "id"
    elif index == "hash":
        return "molecule_hash"
    elif index in ["id", "molecule_hash"]:
        return index
    elif index == "molecular_formula":
        return index
    else:
        raise KeyError("Molecule Index '{}' not understood".format(index))


def translate_generic_index(index):
    if index in ["_id", "id", "ids"]:
        return "_id"
    elif index in ["key"]:
        return "key"
    else:
        raise KeyError("Generic Index '{}' not understood".format(index))


def get_metadata():
    """
    Returns a copy of the metadata for database getters
    """
    return json.loads(_get_metadata)


def add_metadata():
    """
    Returns a copy of the metadata for database save/updates
    """
    return json.loads(_add_metadata)



"""
A ORM for Optimization results
"""

import json


class OptimizationORM:
    """
    A interface to the raw JSON data of a Optimization result.
    """

    # Maps {internal_status : FractalServer status}
    __json_mapper = {
        "_id": "id",
        "_success": "success",
        "hash_index": "hash_index",

        # Keywords
        "_program": "program",
        "_qc_options": "qc_meta",
        "_initial_molecule_id": "initial_molecule",
        "_final_molecule_id": "final_molecule",
        "_trajectory": "trajectory",
        "_energies": "energies",
    }

    def __init__(self, initial_molecule, **kwargs):
        """Initializes a OptimizationORM object, from local data.

        This object may be able to submit tasks to the server in the future.

        *Prototype object, may change in the future.

        Parameters
        ----------
        initial_molecule : TYPE
            Description
        kwargs:
            See OptimizationORM.from_json

        """
        self._initial_molecule = initial_molecule
        self._client = kwargs.pop("client", None)

        # Set kwargs
        for k in self.__json_mapper.keys():
            setattr(self, k, kwargs.get(k[1:], None))

    @classmethod
    def from_json(cls, data, client=None):
        """
        Creates a OptimizationORM object from FractalServer data.

        Parameters
        ----------
        data : dict
            A JSON blob from FractalServer:
                - "id": The service id of the blob
                - "success": If the optimization result was successful or not.
                - "program": The program used for the optimization run.
                - "qc_meta": The quantum chemistry options identified..
                - "initial_molecule_id": The id of the initial (submitted) molecule.
                - "final_molecule_id": The id of the optimizated molecule.
                - "trajectory": QC results for each step in the geometry optimization.
                - "energies": The final energies for each step in the geometry optimization.
        client : FractalClient, optional
            A activate server connection.

        Returns
        -------
        optimization_obj : OptimizationORM
            A OptimizationORM object from the specified JSON.

        """
        kwargs = {}
        for k, v in OptimizationORM.__json_mapper.items():
            if v in data:
                kwargs[k[1:]] = data[v]

        if ("final_energies" in kwargs) and (kwargs["final_energies"] is not None):
            kwargs["final_energies"] = {tuple(json.loads(k)): v for k, v in kwargs["final_energies"].items()}

        kwargs["client"] = client
        return cls(None, **kwargs)

    def __str__(self):
        """
        Simplified optimization string representation.

        Returns
        -------
        ret : str
            A representation of the current Optimization status.

        Examples
        --------

        >>> repr(optimization_obj)
        Optimization(id='5b7f1fd57b87872d2c5d0a6d', status='FINISHED', molecule_id='5b7f1fd57b87872d2c5d0a6c', molecule_name='HOOH')
        """

        ret = "Optimization("
        ret += "id='{}', ".format(self._id)
        ret += "success='{}', ".format(self._success)
        ret += "initial_molecule_id='{}', ".format(self._initial_molecule_id)

        name = None
        if self._initial_molecule:
            name = self._initial_molecule.name()

        ret += "initial_molecule_name='{}')".format(name)

        return ret

    def energies(self):
        """A list of energies along the trajectory path.

        Returns
        -------
        list of float
            The energy of each point in [Eh]
        """
        return self._energies[:]

    def final_energy(self):
        """The final energy of the geometry optimization.

        Returns
        -------
        float
            The optimization molecular energy.
        """
        return self._energies[-1]

    def get_trajectory(self, projection=None):
        """Returns the raw documents for each gradient evaluation in the trajectory.

        Parameters
        ----------
        client : qcportal.FractalClient
            A active client connected to a server.
        projection : None, optional
            A dictionary of the project to apply to the document

        Returns
        -------
        list of dict
            A list of results documents
        """

        return self._client.get_results(id=self._trajectory)

    def final_molecule(self):
        """Returns the optimized molecule

        Returns
        -------
        Molecule
            The optimized molecule
        """

        ret = self._client.get_molecules([self._final_molecule_id], index="id")
        return ret[0]



"""Generic Collection class
"""

from typing import Dict, Any

from .collection import Collection
from .collection_utils import register_collection


class Generic(Collection):
    """
    This is a QCA GenericCollection class. This class behaves similarly to a dictionary, but
    can be serialized and saved to a QCFractal server.

    Attributes
    ----------
    client : client.FractalClient
        A optional server portal to connect the database
    """

    class DataModel(Collection.DataModel):
        """
        Internal Data structure base model typed by PyDantic

        This structure validates input, allows server-side validation and data security,
        and will create the information to pass back and forth between server and client
        """
        data: Dict[str, Any] = {}

        class Config:
            allow_extra = False

    def _pre_save_prep(self, client):
        pass

    def __setitem__(self, key, item):
        self.data.data[key] = item

    def __getitem__(self, key):
        return self.data.data[key]

    def get_data(self, copy: bool=True):
        """Returns a copy of the underlying data object.

        Parameters
        ----------
        copy : bool, optional
            Wether to copy the object or not.

        Returns
        -------
        DataModel
            The underly DataModel
        """
        if copy:
            return self.data.copy(deep=True)
        else:
            return self.data


register_collection(Generic)



"""
DQM Client base folder
"""

from . import collections
from . import data
from . import dict_utils
from . import orm
from . import schema
from .client import FractalClient

# Add imports here
from .models.common_models import Molecule



"""
Wraps geometric procedures
"""

import copy
import json
from typing import Any, Dict, List

import numpy as np

try:
    import torsiondrive
    from torsiondrive import td_api
except ImportError:
    td_api = None

from qcfractal.interface.models.torsiondrive import TorsionDrive
from qcfractal.interface.models.common_models import json_encoders

from .service_util import BaseService, TaskManager

__all__ = ["TorsionDriveService"]


def _check_td():
    if td_api is None:
        raise ImportError("Unable to find TorsionDrive which must be installed to use the TorsionDriveService")


class TorsionDriveService(BaseService):

    # Index info
    status: str = "READY"
    service: str = "torsiondrive"
    program: str = "torsiondrive"
    procedure: str = "torsiondrive"

    # Output
    output: TorsionDrive

    # Temporaries
    torsiondrive_state: Dict[str, Any]
    optimization_history: Dict[str, List[str]] = {}

    # Task helpers
    task_map: Dict[str, List[str]] = {}
    task_manager: TaskManager = TaskManager()

    # Templates
    dihedral_template: str
    optimization_template: str
    molecule_template: str

    class Config:
        json_encoders = json_encoders

    @classmethod
    def initialize_from_api(cls, storage_socket, service_input):
        _check_td()

        # Build the results object
        input_dict = service_input.dict()
        input_dict["initial_molecule"] = [x["id"] for x in input_dict["initial_molecule"]]

        # Validate input
        output = TorsionDrive(
            **input_dict,
            provenance={
                "creator": "torsiondrive",
                "version": torsiondrive.__version__,
                "routine": "torsiondrive.td_api"
            },
            final_energy_dict={},
            minimum_positions={},
            optimization_history={})

        meta = {"output": output}

        # Remove identity info from molecule template
        molecule_template = copy.deepcopy(service_input.initial_molecule[0].json_dict())
        molecule_template.pop("id", None)
        molecule_template.pop("identifiers", None)
        meta["molecule_template"] = json.dumps(molecule_template)

        # Initiate torsiondrive meta
        meta["torsiondrive_state"] = td_api.create_initial_state(
            dihedrals=output.torsiondrive_meta.dihedrals,
            grid_spacing=output.torsiondrive_meta.grid_spacing,
            elements=molecule_template["symbols"],
            init_coords=[x.geometry for x in service_input.initial_molecule])

        # Build dihedral template
        dihedral_template = []
        for idx in output.torsiondrive_meta.dihedrals:
            tmp = {"type": "dihedral", "indices": idx}
            dihedral_template.append(tmp)

        meta["dihedral_template"] = json.dumps(dihedral_template)

        # Build optimization template
        meta["optimization_template"] = json.dumps({
            "meta": {
                "procedure": "optimization",
                "keywords": {"program": output.optimization_meta.program, "values": output.optimization_meta.dict()},
                "program": output.optimization_meta.program,
                "qc_meta": output.qc_meta.dict(),
                "tag": meta.pop("tag", None)
            },
        })

        # Move around geometric data
        meta["optimization_program"] = output.optimization_meta.program

        meta["hash_index"] = output.get_hash_index()

        return cls(**meta, storage_socket=storage_socket)

    def iterate(self):

        self.status = "RUNNING"

        # Check if tasks are done
        print("Done", self.task_manager.done(self.storage_socket))
        if self.task_manager.done(self.storage_socket) is False:
            return False

        complete_tasks = self.task_manager.get_tasks(self.storage_socket)

        # Populate task results
        task_results = {}
        for key, task_ids in self.task_map.items():
            task_results[key] = []

            # Check for history key
            if key not in self.optimization_history:
                self.optimization_history[key] = []

            for task_id in task_ids:
                # Cycle through all tasks for this entry
                ret = complete_tasks[task_id]

                # Lookup molecules
                mol_keys = self.storage_socket.get_molecules(
                    [ret["initial_molecule"], ret["final_molecule"]], index="id")["data"]

                task_results[key].append((mol_keys[0]["geometry"], mol_keys[1]["geometry"], ret["energies"][-1]))

                # Update history
                self.optimization_history[key].append(ret["id"])

        td_api.update_state(self.torsiondrive_state, task_results)

        # Create new tasks from the current state
        next_tasks = td_api.next_jobs_from_state(self.torsiondrive_state, verbose=True)

        # All done
        if len(next_tasks) == 0:
            return self.finalize()

        self.submit_optimization_tasks(next_tasks)

        return False

    def submit_optimization_tasks(self, task_dict):

        new_tasks = {}
        task_map = {}

        for key, geoms in task_dict.items():
            task_map[key] = []
            for num, geom in enumerate(geoms):

                # Update molecule
                packet = json.loads(self.optimization_template)

                # Construct constraints
                constraints = json.loads(self.dihedral_template)
                grid_id = td_api.grid_id_from_string(key)
                for con_num, k in enumerate(grid_id):
                    constraints[con_num]["value"] = k
                packet["meta"]["keywords"]["values"]["constraints"] = {"set": constraints}

                # Build new molecule
                mol = json.loads(self.molecule_template)
                mol["geometry"] = geom
                packet["data"] = [mol]

                task_key = "{}-{}".format(key, num)
                new_tasks[task_key] = packet

                task_map[key].append(task_key)

        self.task_manager.submit_tasks(self.storage_socket, "optimization", new_tasks)
        self.task_map = task_map

    def finalize(self):
        """
        Finishes adding data to the TorsionDrive object
        """

        self.output.Config.allow_mutation = True
        self.output.success = True
        self.output.status = "COMPLETE"

        # # Get lowest energies and positions
        for k, v in self.torsiondrive_state["grid_status"].items():
            min_pos = int(np.argmin([x[2] for x in v]))
            key = json.dumps(td_api.grid_id_from_string(k))
            self.output.minimum_positions[key] = min_pos
            self.output.final_energy_dict[key] = v[min_pos][2]

        self.output.optimization_history = {
            json.dumps(td_api.grid_id_from_string(k)): v
            for k, v in self.optimization_history.items()
        }

        self.output.Config.allow_mutation = False
        return self.output



"""
A model for GridOptimization
"""

import json
from typing import Any, Dict, List, Tuple, Union

from pydantic import BaseModel, validator

from .common_models import QCMeta, Provenance, Molecule, json_encoders, hash_dictionary

__all__ = ["GridOptimizationInput", "GridOptimization"]


class ScanDimension(BaseModel):
    """
    A dimension to scan over
    """
    type: str
    indices: List[int]
    steps: List[float]

    @validator('type')
    def check_type(cls, v, values, **kwargs):
        possibilities = {"distance", "angle", "dihedral"}
        if v not in possibilities:
            raise TypeError("Type '{}' found, can only be one of {}.".format(v, possibilities))

        return v

    @validator('indices', whole=True)
    def check_indices(cls, v, values, **kwargs):
        sizes = {"distance": 2, "angle": 3, "dihedral": 4}
        if sizes[values["type"]] != len(v):
            raise ValueError("ScanDimension of type {} must have {} values, found {}.".format(
                values["type"], sizes[values["type"]], len(v)))

        return v

    @validator('steps', whole=True)
    def check_steps_monotonic(cls, v):
        if not (all(x < y for x, y in zip(v, v[1:])) or all(x > y for x, y in zip(v, v[1:]))):
            raise ValueError("Steps are not monotonically increasing or decreasing.")

        return v


class GOOptions(BaseModel):
    """
    GridOptimization options
    """
    scans: List[ScanDimension]
    starting_grid: str = "relative"

    @validator('starting_grid')
    def check_starting_grid(cls, v):
        v = v.lower()
        if v not in ["relative", "zero"]:
            raise KeyError("Starting grid can only be relative or zero, found '{}'".found(v))

        return v

    class Config:
        allow_mutation = False


class OptOptions(BaseModel):
    """
    GridOptimization options
    """
    program: str

    class Config:
        allow_extra = True
        allow_mutation = False


class GridOptimizationInput(BaseModel):
    """
    A GridOptimization Input base class
    """

    program: str = "qcfractal"
    procedure: str = "gridoptimization"
    initial_molecule: Union[str, Molecule]
    gridoptimization_meta: GOOptions
    optimization_meta: OptOptions
    qc_meta: QCMeta

    class Config:
        allow_mutation = False
        json_encoders = json_encoders

    def get_hash_index(self):
        if isinstance(self.initial_molecule, str):
            mol_id = self.initial_molecule

        else:
            if self.initial_molecule.id is None:
                raise ValueError("Cannot get the hash_index without a valid intial_molecule.id field.")

            mol_id = self.initial_molecule.id

        data = self.dict(include=["program", "procedure", "gridoptimization_meta", "optimization_meta", "qc_meta"])
        data["initial_molecule"] = mol_id

        return hash_dictionary(data)

    def serialize_key(self, key):
        if isinstance(key, (int, float)):
            key = (int(key), )

        return json.dumps(key)

    def deserialize_key(self, key):

        return tuple(json.loads(key))

    def get_scan_value(self, key: Union[str, Tuple]) -> Tuple:
        """
        Obtains the scan parameters at a given grid point.
        """
        if isinstance(key, str):
            key = self.deserialize_key(key)

        ret = []
        for n, idx in enumerate(key):
            ret.append(self.gridoptimization_meta.scans[n].steps[idx])

        return tuple(ret)

    def get_scan_dimensions(self) -> Tuple:
        """
        Returns the overall dimensions of the scan.
        """
        ret = []
        for scan in self.gridoptimization_meta.scans:
            ret.append(len(scan.steps))

        return tuple(ret)


class GridOptimization(GridOptimizationInput):
    """
    A interface to the raw JSON data of a GridOptimization torsion scan run.
    """

    # Client and local data
    client: Any = None
    cache: Dict[str, Any] = {}

    # Identification
    id: str = None
    success: bool = False
    status: str = "INCOMPLETE"
    hash_index: str = None

    provenance: Provenance

    # Data pointers
    initial_molecule: str
    final_energy_dict: Dict[str, float]
    grid_optimizations: Dict[str, str]
    starting_grid: tuple

    class Config:
        allow_mutation = False
        json_encoders = json_encoders

    def __init__(self, **data):
        super().__init__(**data)

        # Set hash index if not present
        if self.hash_index is None:
            self.__values__["hash_index"] = self.get_hash_index()

    def __str__(self):
        """
        Simplified gridoptimization string representation.

        Returns
        -------
        ret : str
            A representation of the current GridOptimization status.

        Examples
        --------

        >>> repr(torsiondrive_obj)
        GridOptimization(id='5b7f1fd57b87872d2c5d0a6d', success=True, molecule_id='5b7f1fd57b87872d2c5d0a6c')
        """

        ret = "GridOptimization("
        ret += "id='{}', ".format(self.id)
        ret += "success='{}', ".format(self.success)
        ret += "initial_molecule='{}')".format(self.initial_molecule)

        return ret

## Utility

    def dict(self, exclude=set(), include=None, by_alias=False):
        exclude |= {"client", "cache"}
        return super().dict(include=include, exclude=exclude, by_alias=by_alias)

    def json_dict(self):
        return json.loads(self.json())

## Query

    def final_energies(self, key=None):

        if key is None:
            return {self.deserialize_key(k): v for k, v in self.final_energy_dict.items()}
        else:
            return self.final_energy_dict[self.serialize_key(key)]



"""
Mongo QCDB Abstract basic Collection class

Helper
"""

import abc
import copy
import json
from typing import Dict, Optional

from pydantic import BaseModel


class Collection(abc.ABC):
    def __init__(self, name: str, **kwargs):
        """
        Initializer for the Collections objects. If no Portal is supplied or the Collection name
        is not present on the server that the Portal is connected to a blank Collection will be
        created.

        Parameters
        ----------
        name : str
            The name of the Collection object as ID'ed on the storage backend.
        client : client.FractalClient, optional
            A Portal client to connect to a server
        **kwargs
            Additional keywords which are passed to the Collection and the initial data constructor
            It is up to the individual implementations of the Collection to do things with that data
        """

        self.client = kwargs.pop("client", None)
        if (self.client is not None) and not (self.client.__class__.__name__ == "FractalClient"):
            raise TypeError("Expected FractalClient as `client` kwarg, found {}.".format(type(self.client)))

        if 'collection' not in kwargs:
            kwargs['collection'] = self.__class__.__name__.lower()

        kwargs['name'] = name

        # Create the data model
        self.data = self.DataModel(**kwargs)

    class DataModel(BaseModel):
        """
        Internal Data structure base model typed by PyDantic

        This structure validates input, allows server-side validation and data security,
        and will create the information to pass back and forth between server and client

        Subclasses of Collection can extend this class internally to change the set of
        additional data defined by the Collection
        """
        name: str
        collection: str = None
        provenance: Dict[str, str] = {}
        id: str = 'local'

        class Config:
            # Allows extra args to prevent subclass errors, but just ignore them
            # These might be the defaults, but make it explicit to be sure
            allow_extra = False
            ignore_extra = True

    def __str__(self) -> str:
        """
        A simple string representation of the Collection.

        Returns
        -------
        ret : str
            A representation of the Collection.

        Examples
        --------

        >>> repr(obj)
        Collection(name=`S22`, id='5b7f1fd57b87872d2c5d0a6d', client=`localhost:8888`)
        """

        client = None
        if self.client:
            client = self.client.address

        class_name = self.__class__.__name__
        ret = "{}(".format(class_name)
        ret += "name=`{}`, ".format(self.data.name)
        ret += "id='{}', ".format(self.data.id)
        ret += "client='{}') ".format(client)

        return ret

    @classmethod
    def from_server(cls, client, name):
        """Creates a new class from a server

        Parameters
        ----------
        client : client.FractalClient
            A Portal client to connected to a server
        name : str
            The name of the collection to pull from.

        Returns
        -------
        Collection
            A ODM of the data.
        """

        if not (client.__class__.__name__ == "FractalClient"):
            raise TypeError("Expected a FractalClient as first argument, found {}.".format(type(client)))

        class_name = cls.__name__.lower()
        tmp_data = client.get_collection(class_name, name, full_return=True)
        if tmp_data.meta.n_found == 0:
            raise KeyError("Warning! `{}: {}` not found.".format(class_name, name))

        return cls.from_json(tmp_data.data[0], client=client)

    @classmethod
    def from_json(cls, data, client=None):
        """Creates a new class from a JSON blob

        Parameters
        ----------
        data : dict
            The JSON blob to create a new class from.
        client : client.FractalClient
            A Portal client to connected to a server

        Returns
        -------
        Collection
            A ODM of the data.

        """
        # Check we are building the correct object
        class_name = cls.__name__.lower()
        if "collection" not in data:
            raise KeyError("Attempted to create Collection from JSON, but no `collection` field found.")

        if data["collection"] != class_name:
            raise KeyError("Attempted to create Collection from JSON with class {}, but found collection type of {}.".
                           format(class_name, data["collection"]))

        # Attempt to build class
        # First make sure external source provides ALL keys, including "optional" ones
        # Assumption here is that incoming source should have all this from a previous instance
        # and will not have missing fields. Also enforces that no default values are *assumed* from
        # external input.
        # PyDantic can only enforce required on init entries, but lets everything else have defaults,
        # this check asserts all fields are present, even if their default values are chosen
        # Also provides consistency check in case defaults ever change in the future.
        # This check could be removed though without any code failures
        req_fields = cls.DataModel.__fields__.keys()
        missing = req_fields - data.keys()
        if len(missing):
            raise KeyError("For class {} the following fields are missing {}.".format(class_name, missing))
        name = data.pop('name')
        # Allow PyDantic to handle type validation
        return cls(name, client=client, **data)

    def to_json(self, filename: Optional[str]=None):
        """
        If a filename is provided, dumps the file to disk. Otherwise returns a copy of the current data.

        Parameters
        ----------
        filename : str, Optional, Default: None
            The filename to drop the data to.

        Returns
        -------
        ret : dict
            A JSON representation of the Collection
        """
        data = self.data.dict()
        if filename is not None:
            with open(filename, 'w') as open_file:
                json.dump(data, open_file)
        else:
            return copy.deepcopy(data)

    @abc.abstractmethod
    def _pre_save_prep(self, client):
        """
        Additional actions to take before saving, done as the last step before data is written.

        This does not return anything but can prep the `self.data` field before storing it.

        Has access to the `client` in case its needed to do pre-conditioning.

        Parameters
        ----------
        client : FractalClient
            Client to use for storage access
        """
        pass

    # Setters
    def save(self, client=None, overwrite: bool=False):
        """Uploads the overall structure of the Collection (indices, options, new molecules, etc)
        to the server.

        Parameters
        ----------
        client : None, optional
            A Portal object to the server to upload to
        overwrite : bool, optional
            Overwrite the data in the server on not

        """
        class_name = self.__class__.__name__.lower()
        if self.data.name == "":
            raise AttributeError("Collection:save: {} must have a name!".format(class_name))

        if client is None:
            if self.client is None:
                raise AttributeError("Collection:save: {} does not own a Storage Database "
                                     "instance and one was not passed in.".format(class_name))
            client = self.client

        if overwrite and (self.data.id == self.data.fields['id'].default):
            raise KeyError("Attempting to overwrite the {} class on the server, but no ID found.".format(class_name))

        self._pre_save_prep(client)

        # Add the database
        return client.add_collection(self.data.dict(), overwrite=overwrite)



"""
A command line interface to the qcfractal server.
"""

import argparse

import tornado.log

import qcfractal
from . import cli_utils

__all__ = ["main"]


def parse_args():
    parser = argparse.ArgumentParser(description='A CLI for the QCFractal QueueManager.')
    subparsers = parser.add_subparsers(help='QueueManager Backend Type', dest='adapter_type')
    subparsers.required = True

    # Keywords for Dask
    dask_parser = subparsers.add_parser('dask', help='Dask')
    dask_parser.add_argument("--dask-uri", type=str, help="URI of the dask-server")
    dask_parser.add_argument(
        "--local-cluster", action="store_true", help="Start a Dask LocalCluster rather than connect to a scheduler")
    dask_parser.add_argument(
        "--local-workers", type=int, default=None, help="The number of workers for the LocalCluster")

    # Keywords for Fireworks
    fw_parser = subparsers.add_parser('fireworks', help='Fireworks')
    fw_parser.add_argument("--fw-config", type=str, help="A FWConfig file")
    fw_parser.add_argument("--fw-uri", type=str, help="URI of MongoDB server")
    fw_parser.add_argument(
        "--fw-name", type=str, default="qcfractal_fireworks_manager", help="The MongoDB Database to use locally")

    # Keywords for ProcessPoolExecutor
    executor_parser = subparsers.add_parser('executor', help='ProcessPoolExecutor')
    executor_parser.add_argument("--nprocs", type=int, help="The number of process for the executor")

    # FractalClient options
    parser.add_argument(
        "--fractal-uri", type=str, default="localhost:7777", help="FractalServer location to pull from")
    parser.add_argument("-u", "--username", type=str, help="FractalServer username")
    parser.add_argument("-p", "--password", type=str, help="FractalServer password")
    parser.add_argument("--noverify", action="store_true", default=True, help="The logfile prefix to use")

    # QueueManager options
    parser.add_argument(
        "--max-tasks", type=int, default=1000, help="Maximum number of tasks to hold at any given time.")
    parser.add_argument("--cluster-name", type=str, default="unknown", help="The name of the compute cluster to start")
    parser.add_argument("--queue-tag", type=str, help="The queue tag to pull from")
    parser.add_argument("--logfile-prefix", type=str, default=None, help="The prefix of the logfile to write to.")
    parser.add_argument(
        "--update-frequency", type=int, default=15, help="The frequency in seconds to check for complete tasks.")

    # Additional args
    parser.add_argument("--test", action="store_true", help="Boot and run a short test suite to validate setup")
    parser.add_argument("--rapidfire", action="store_true", help="Boot and run jobs until complete")
    parser.add_argument("--config-file", type=str, default=None, help="A configuration file to use")
    args = vars(parser.parse_args())
    if args["config_file"] is not None:
        data = cli_utils.read_config_file(args["config_file"])
        args = cli_utils.argparse_config_merge(parser, args, data, parser_default=[args["adapter_type"]])

    if args["rapidfire"] and args["test"]:
        raise KeyError("Can only either select 'rapidfire' or 'test' options, but not both.")

    return args


def main(args=None):

    # Grab CLI args if not present
    if args is None:
        args = parse_args()

    exit_callbacks = []

    # Handle Dask adapters
    if args["adapter_type"] == "dask":
        dd = cli_utils.import_module("distributed")

        if args["local_cluster"]:
            # Build localcluster and exit callbacks
            queue_client = dd.Client(threads_per_worker=1, n_workers=args["local_workers"])
        else:
            if args["dask_uri"] is None:
                raise KeyError("A 'dask-uri' must be specified.")
            queue_client = dd.Client(args["dask_uri"])

    # Handle Fireworks adapters
    elif args["adapter_type"] == "fireworks":

        # Check option conflicts
        num_options = sum(args[x] is not None for x in ["fw_config", "fw_uri"])
        if num_options == 0:
            args["fw_uri"] = "mongodb://localhost:27017"
        elif num_options != 1:
            raise KeyError("Can only provide a single URI or config_file for Fireworks.")

        fireworks = cli_utils.import_module("fireworks")

        if args["fw_uri"] is not None:
            queue_client = fireworks.LaunchPad(args["fw_uri"], name=args["fw_name"])
        elif args["fw_config"] is not None:
            queue_client = fireworks.LaunchPad.from_file(args["fw_config"])
        else:
            raise KeyError("A URI or config_file must be specified.")

    elif args["adapter_type"] == "executor":

        from concurrent.futures import ProcessPoolExecutor

        queue_client = ProcessPoolExecutor(max_workers=args["nprocs"])

    else:
        raise KeyError(
            "Unknown adapter type '{}', available options: 'fireworks', 'dask', .".format(args["adapter_type"]))

    # Quick logging
    if args["logfile_prefix"] is not None:
        tornado.options.options['log_file_prefix'] = logfile_prefix
    tornado.log.enable_pretty_logging()

    # Build the client
    if args["test"]:
        client = None
    else:
        client = qcfractal.interface.FractalClient(
            args["fractal_uri"], username=args["username"], password=args["password"], verify=(not args["noverify"]))

    # Build out the manager itself
    manager = qcfractal.queue.QueueManager(
        client,
        queue_client,
        max_tasks=args["max_tasks"],
        queue_tag=args["queue_tag"],
        cluster=args["cluster_name"],
        update_frequency=args["update_frequency"])

    # Add exit callbacks
    for cb in exit_callbacks:
        manager.add_exit_callback(cb[0], *cb[1], **cb[2])

    # Either startup the manager or run until complete
    if args["rapidfire"]:
        manager.await_results()
    elif args["test"]:
        success = manager.test()
        if success is False:
            raise ValueError("Testing was not successful, failing.")
    else:

        cli_utils.install_signal_handlers(manager.loop, manager.stop)

        # Blocks until keyboard interupt
        manager.start()


if __name__ == '__main__':
    main()
